# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Base class and star finder kernel for detecting stars in an astronomical
image.

Each star-finding class should define a method called ``find_stars``
that finds stars in an image.
"""

import abc
import inspect
import math
import warnings

import astropy.units as u
import numpy as np
from astropy.stats import gaussian_fwhm_to_sigma
from astropy.utils import lazyproperty
from astropy.utils.exceptions import AstropyDeprecationWarning

from photutils.detection.peakfinder import find_peaks
from photutils.utils._deprecation import (create_empty_deprecated_qtable,
                                          deprecated_getattr,
                                          deprecated_positional_kwargs,
                                          deprecated_renamed_argument)
from photutils.utils._misc import _get_meta
from photutils.utils._quantity_helpers import check_units
from photutils.utils._repr import make_repr
from photutils.utils.cutouts import _make_cutouts
from photutils.utils.exceptions import NoDetectionsWarning

__all__ = ['StarFinderBase', 'StarFinderCatalogBase']

# Remove in 4.0
_DEPRECATED_ATTRIBUTES: dict = {
    'xcentroid': 'x_centroid',
    'ycentroid': 'y_centroid',
    'cutout_xcentroid': 'cutout_x_centroid',
    'cutout_ycentroid': 'cutout_y_centroid',
    'pa': 'orientation',
    'npix': 'n_pixels',
}


class StarFinderBase(metaclass=abc.ABCMeta):
    """
    Abstract base class for star finders.
    """

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __call__(self, data, mask=None):
        """
        Find stars in an astronomical image.

        Parameters
        ----------
        data : 2D array_like
            The 2D image array.

        mask : 2D bool array, optional
            A boolean mask with the same shape as ``data``, where a
            `True` value indicates the corresponding element of ``data``
            is masked. Masked pixels are ignored when searching for
            stars.

        Returns
        -------
        table : `~astropy.table.Table` or `None`
            A table of found stars. If no stars are found then `None` is
            returned.
        """
        return self.find_stars(data, mask=mask)

    @staticmethod
    def _find_stars(convolved_data, kernel, threshold, *, min_separation=0.0,
                    mask=None, exclude_border=False):
        """
        Find stars in an image.

        Parameters
        ----------
        convolved_data : 2D array_like
            The convolved 2D array. Should be NaN-free; any NaN values
            should be handled before calling this method.

        kernel : `_StarFinderKernel` or 2D `~numpy.ndarray`
            The convolution kernel. ``StarFinder`` inputs the kernel
            as a 2D array.

        threshold : float or 2D array_like
            The absolute image value above which to select sources. The
            exact value depends on the calling star finder class (e.g.,
            `DAOStarFinder` multiplies the ``threshold`` by the kernel
            relative error, whereas `IRAFStarFinder` and `StarFinder`
            directly use the input ``threshold``). A 2D ``threshold``
            must have the same shape as ``convolved_data``. If
            ``convolved_data`` is a `~astropy.units.Quantity` array,
            then ``threshold`` must have the same units.

        min_separation : float, optional
            The minimum separation for detected objects in pixels.

        mask : 2D bool array, optional
            A boolean mask with the same shape as ``data``, where a
            `True` value indicates the corresponding element of ``data``
            is masked. Masked pixels are ignored when searching for
            stars.

        exclude_border : bool, optional
            Set to `True` to exclude sources found within half the
            size of the convolution kernel from the image borders.
            The default is `False`, which is the mode used by IRAF's
            `DAOFIND
            <https://iraf.readthedocs.io/en/latest/tasks/noao/digiphot/apphot/daofind.html>`_
            and `STARFIND
            <https://iraf.readthedocs.io/en/latest/tasks/images/imcoords/starfind.html>`_.

        Returns
        -------
        result : Nx2 `~numpy.ndarray` or `None`
            An Nx2 array containing the (x, y) pixel coordinates. `None`
            is returned if no sources are found.
        """
        # Define a local footprint for the peak finder
        find_peaks_kwargs = {}
        if min_separation == 0:  # use kernel-shape footprint
            if isinstance(kernel, np.ndarray):
                footprint = np.ones(kernel.shape)
            else:
                footprint = kernel.mask.astype(bool)
            find_peaks_kwargs['footprint'] = footprint
        else:
            find_peaks_kwargs['min_separation'] = min_separation

        # Define the border exclusion region
        if exclude_border:
            if isinstance(kernel, np.ndarray):
                yborder = (kernel.shape[0] - 1) // 2
                xborder = (kernel.shape[1] - 1) // 2
            else:
                yborder = kernel.y_radius
                xborder = kernel.x_radius
            border_width = (yborder, xborder)
        else:
            border_width = None

        # Find local peaks in the convolved data.
        # Suppress any NoDetectionsWarning from find_peaks.
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=NoDetectionsWarning)
            tbl = find_peaks(convolved_data, threshold, mask=mask,
                             border_width=border_width, **find_peaks_kwargs)

        if tbl is None:
            return None

        return np.transpose((tbl['x_peak'], tbl['y_peak']))

    @abc.abstractmethod
    def find_stars(self, data, *, mask=None):
        """
        Find stars in an astronomical image.

        Parameters
        ----------
        data : 2D array_like
            The 2D image array.

        mask : 2D bool array, optional
            A boolean mask with the same shape as ``data``, where a
            `True` value indicates the corresponding element of ``data``
            is masked. Masked pixels are ignored when searching for
            stars.

        Returns
        -------
        table : `~astropy.table.Table` or `None`
            A table of found stars. If no stars are found then `None` is
            returned.
        """


class StarFinderCatalogBase(metaclass=abc.ABCMeta):
    """
    Abstract base class for star finder catalogs.

    This class provides common functionality for catalog classes that
    store and compute properties of detected sources. External packages
    may subclass it to create custom star finder catalogs.

    Subclasses **must** implement:

    * :attr:`x_centroid` property -- Object centroid in the x direction.
    * :attr:`y_centroid` property -- Object centroid in the y direction.
    * `apply_filters` method -- Filter the catalog using
      algorithm-specific criteria.
    * ``default_columns`` attribute -- A tuple of column names used
      by `to_table` when no explicit columns are given. This should
      be set in the subclass ``__init__``.

    Subclasses **may** override:

    * `_get_init_attributes` -- Return attribute names to copy
      during slicing. The override should include
      ``'default_columns'`` in the returned tuple.
    * `make_cutouts` -- Customize how cutout arrays are extracted.
    * `cutout_data` -- Customize the cutouts used for photometry
      (e.g., zeroing negative pixels).

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The 2D image. The image should be background-subtracted.

    xypos : Nx2 `~numpy.ndarray`
        An Nx2 array of (x, y) pixel coordinates denoting the central
        positions of the stars.

    kernel : 2D `~numpy.ndarray`
        A 2D array of the PSF kernel. Internally, the star finder
        classes may also pass a kernel object.

    n_brightest : int, None, optional
        The number of brightest objects to keep after sorting the source
        list by flux. If ``n_brightest`` is set to `None`, all objects
        will be selected.

    peak_max : float, None, optional
        The maximum allowed peak pixel value in an object. Objects with
        peak pixel values greater than ``peak_max`` will be rejected.
        This keyword may be used, for example, to exclude saturated
        sources. If the star finder is run on an image that is a
        `~astropy.units.Quantity` array, then ``peak_max`` must have the
        same units. If ``peak_max`` is set to `None`, then no peak pixel
        value filtering will be performed.
    """

    @deprecated_renamed_argument('brightest', 'n_brightest', '3.0',
                                 until='4.0')
    @deprecated_renamed_argument('peakmax', 'peak_max', '3.0', until='4.0')
    def __init__(self, data, xypos, kernel, *, n_brightest=None,
                 peak_max=None):
        # Validate the units
        check_units((data, peak_max), ('data', 'peak_max'))

        self.data = data
        unit = data.unit if isinstance(data, u.Quantity) else None
        self.unit = unit
        self.kernel = kernel
        self.cutout_shape = kernel.shape

        self.xypos = np.atleast_2d(xypos)
        self.n_brightest = n_brightest
        self.peak_max = peak_max
        self.default_columns = ()

        self.id = np.arange(len(self)) + 1

    def __repr__(self):
        params = ('nsources',)
        overrides = {'nsources': len(self)}
        return make_repr(self, params, brackets=True, overrides=overrides)

    def __str__(self):
        params = ('nsources',)
        overrides = {'nsources': len(self)}
        return make_repr(self, params, overrides=overrides, long=True)

    def __len__(self):
        return len(self.xypos)

    def __getitem__(self, index):
        """
        Index or slice the catalog.

        This method should be overridden in subclasses to handle
        class-specific attributes.
        """
        # NOTE: we allow indexing/slicing of scalar (self.isscalar = True)
        #       instances in order to perform catalog filtering even for
        #       a single source

        newcls = object.__new__(self.__class__)

        # Get attributes to copy from subclass
        init_attr = self._get_init_attributes()
        for attr in init_attr:
            setattr(newcls, attr, getattr(self, attr))

        # xypos determines ordering and isscalar
        # NOTE: always keep as 2D array, even for a single source
        attr = 'xypos'
        value = getattr(self, attr)[index]
        setattr(newcls, attr, np.atleast_2d(value))

        # Index/slice the remaining attributes
        keys = set(self.__dict__.keys()) & set(self._lazyproperties)
        keys.add('id')
        for key in keys:
            value = self.__dict__[key]

            # Do not insert lazy attributes that are always scalar (e.g.,
            # isscalar), i.e., not an array/list for each source
            if np.isscalar(value):
                continue

            # Ensure value is always at least a 1D array, even for a
            # single source
            value = np.atleast_1d(value[index])

            newcls.__dict__[key] = value

        return newcls

    def _get_init_attributes(self):
        """
        Return a tuple of attribute names to copy during slicing.

        This method should be overridden in subclasses.
        """
        return ('data', 'unit', 'kernel', 'n_brightest', 'peak_max',
                'cutout_shape', 'default_columns')

    @property
    def _lazyproperties(self):
        """
        Return all lazyproperties (even in superclasses).

        The result is cached on the class to avoid repeated
        introspection via `inspect.getmembers`.
        """
        cls = self.__class__
        attr = '_cached_lazyproperties'
        # Subclasses get their own lazyproperty list
        if attr not in cls.__dict__:
            def islazyproperty(obj):
                return isinstance(obj, lazyproperty)

            setattr(cls, attr,
                    [i[0] for i in inspect.getmembers(
                        cls, predicate=islazyproperty)])
        return getattr(cls, attr)

    @lazyproperty
    def isscalar(self):
        """
        Whether the instance is scalar (e.g., a single source).
        """
        return self.xypos.shape == (1, 2)

    def reset_ids(self):
        """
        Reset the ID column to be consecutive integers.
        """
        self.id = np.arange(len(self)) + 1

    def make_cutouts(self, data):
        """
        Make cutouts from the data array.

        Parameters
        ----------
        data : 2D `~numpy.ndarray`
            The 2D image array.

        Returns
        -------
        cutouts : 3D `~numpy.ndarray`
            The cutout arrays.
        """
        data_arr = data.value if isinstance(data, u.Quantity) else data

        cutouts, _ = _make_cutouts(data_arr, self.xypos[:, 0],
                                   self.xypos[:, 1], self.cutout_shape)

        if self.unit is not None:
            cutouts <<= self.unit

        return cutouts

    @lazyproperty
    def cutout_data(self):
        """
        The cutout data arrays.

        Subclasses may override this property to customize the cutouts
        used for moment-based photometry calculations (e.g., zeroing
        negative pixels or subtracting a local sky background).
        """
        return self.make_cutouts(self.data)

    @lazyproperty
    def moments(self):
        """
        The raw image moments.
        """
        data = self.cutout_data
        if isinstance(data, u.Quantity):
            data = data.value
        ky, kx = data.shape[1], data.shape[2]
        y = np.arange(ky, dtype=float)
        x = np.arange(kx, dtype=float)
        ypowers = np.column_stack([np.ones(ky), y])  # (ky, 2)
        xpowers = np.column_stack([np.ones(kx), x])  # (kx, 2)
        # M[n, p, q] = sum_jk data[n,j,k] * y[j]^p * x[k]^q
        return ypowers.T @ data @ xpowers

    @lazyproperty
    def moments_central(self):
        """
        The central image moments.
        """
        data = self.cutout_data
        if isinstance(data, u.Quantity):
            data = data.value
        ky, kx = data.shape[1], data.shape[2]
        y = np.arange(ky, dtype=float)
        x = np.arange(kx, dtype=float)
        # Per-source shifted coordinates
        dy = y[np.newaxis, :] - self.cutout_y_centroid[:, np.newaxis]
        dx = x[np.newaxis, :] - self.cutout_x_centroid[:, np.newaxis]
        # Per-source power arrays: (n, ky, 3) and (n, kx, 3)
        ypowers = np.stack([np.ones_like(dy), dy, dy**2], axis=-1)
        xpowers = np.stack([np.ones_like(dx), dx, dx**2], axis=-1)
        # Batched matmul: ypowers^T @ data @ xpowers per source
        moments = (np.transpose(ypowers, (0, 2, 1)) @ data @ xpowers)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            return moments / self.moments[:, 0, 0][:, np.newaxis, np.newaxis]

    @lazyproperty
    def cutout_centroid(self):
        """
        The cutout centroids.
        """
        moments = self.moments

        # Ignore divide-by-zero RuntimeWarning
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            y_centroid = moments[:, 1, 0] / moments[:, 0, 0]
            x_centroid = moments[:, 0, 1] / moments[:, 0, 0]
        return np.transpose((y_centroid, x_centroid))

    @lazyproperty
    def cutout_x_centroid(self):
        """
        The cutout x centroids.
        """
        return np.transpose(self.cutout_centroid)[1]

    @lazyproperty
    def cutout_y_centroid(self):
        """
        The cutout y centroids.
        """
        return np.transpose(self.cutout_centroid)[0]

    @property
    @abc.abstractmethod
    def x_centroid(self):
        """
        Object centroid in the x direction.

        This property must be implemented in subclasses.
        """

    @property
    @abc.abstractmethod
    def y_centroid(self):
        """
        Object centroid in the y direction.

        This property must be implemented in subclasses.
        """

    # Remove in 4.0
    def __getattr__(self, name):
        return deprecated_getattr(self, name, _DEPRECATED_ATTRIBUTES,
                                  since='3.0', until='4.0')

    @lazyproperty
    def mu_sum(self):
        """
        The sum of the central moments.
        """
        return (self.moments_central[:, 0, 2]
                + self.moments_central[:, 2, 0])

    @lazyproperty
    def mu_diff(self):
        """
        The difference of the central moments.
        """
        return (self.moments_central[:, 0, 2]
                - self.moments_central[:, 2, 0])

    @lazyproperty
    def fwhm(self):
        """
        The FWHM of the sources.
        """
        return 2.0 * np.sqrt(np.log(2.0) * self.mu_sum)

    @lazyproperty
    def orientation(self):
        """
        The angle between the ``x`` axis and the major axis of the 2D
        Gaussian function that has the same second-order moments as the
        source.

        The angle increases in the counter-clockwise direction and
        will be in the range [0, 360) degrees.
        """
        angle = 0.5 * np.arctan2(2.0 * self.moments_central[:, 1, 1],
                                 self.mu_diff)
        return (np.rad2deg(angle) % 360) << u.deg

    @lazyproperty
    def roundness(self):
        """
        The roundness of the sources.
        """
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            return (np.sqrt(self.mu_diff**2
                            + 4.0 * self.moments_central[:, 1, 1]**2)
                    / self.mu_sum)

    @lazyproperty
    def peak(self):
        """
        The peak pixel values.
        """
        return np.max(self.cutout_data, axis=(1, 2))

    @lazyproperty
    def flux(self):
        """
        The instrumental fluxes.
        """
        return np.sum(self.cutout_data, axis=(1, 2))

    @lazyproperty
    def mag(self):
        """
        The instrumental magnitudes.
        """
        # Ignore RuntimeWarning if flux is <= 0
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            flux = self.flux
            if isinstance(flux, u.Quantity):
                flux = flux.value
            return -2.5 * np.log10(flux)

    def select_brightest(self):
        """
        Sort the catalog by the brightest fluxes and select the top
        brightest sources.
        """
        newcat = self
        if self.n_brightest is not None:
            idx = np.argsort(self.flux)[::-1][:self.n_brightest]
            newcat = self[idx]
        return newcat

    def _filter_finite(self, attrs, *, initial_mask=None,
                       skip_attrs=()):
        """
        Filter the catalog by removing sources with non-finite values.

        Parameters
        ----------
        attrs : tuple of str
            Attribute names to check for finiteness.

        initial_mask : 1D `~numpy.ndarray` of bool or `None`, optional
            A pre-existing boolean mask to combine with. If `None`,
            starts with all `True`.

        skip_attrs : tuple of str, optional
            Attribute names to skip during finiteness checking.

        Returns
        -------
        catalog : ``self.__class__`` or `None`
            The filtered catalog, or `None` if no sources remain.
        """
        if initial_mask is None:
            mask = np.ones(len(self), dtype=bool)
        else:
            mask = initial_mask.copy()

        for attr in attrs:
            if attr in skip_attrs:
                continue
            mask &= np.isfinite(getattr(self, attr))
        newcat = self[mask]

        if len(newcat) == 0:
            msg = 'No sources were found.'
            warnings.warn(msg, NoDetectionsWarning)
            return None

        return newcat

    def _filter_bounds(self, bounds, *, initial_mask=None, peakattr='peak'):
        """
        Filter the catalog by sharpness, roundness, and peak_max bounds.

        Parameters
        ----------
        bounds : list of tuple
            Each tuple is ``(attr_name, range)`` giving the attribute to
            check and the range of allowed values. The range is a tuple
            of the form ``(lower_bound, upper_bound)``, or `None` to
            skip filtering for that attribute.

        initial_mask : 1D `~numpy.ndarray` of bool or `None`, optional
            A pre-existing boolean mask to combine with. If `None`,
            starts with all `True`.

        peakattr : str, optional
            The attribute name for the peak value used for peak_max
            filtering. The default is ``'peak'``.

        Returns
        -------
        catalog : ``self.__class__`` or `None`
            The filtered catalog, or `None` if no sources remain.
        """
        if initial_mask is None:
            mask = np.ones(len(self), dtype=bool)
        else:
            mask = initial_mask.copy()

        for attr, range_val in bounds:
            if range_val is None:
                continue
            min_val, max_val = range_val
            values = getattr(self, attr)
            mask &= (values >= min_val)
            mask &= (values <= max_val)

        # peak_max filtering is applied separately from the bounds list
        # because it uses a different attribute (peakattr) and is always
        # a single upper bound, not a range.
        if self.peak_max is not None:
            mask &= (getattr(self, peakattr) <= self.peak_max)

        newcat = self[mask]

        if len(newcat) == 0:
            msg = 'Sources were found, but none pass the filtering criteria'
            warnings.warn(msg, NoDetectionsWarning)
            return None

        return newcat

    @abc.abstractmethod
    def apply_filters(self):
        """
        Filter the catalog.

        This method must be implemented in subclasses to apply
        algorithm-specific filtering criteria.
        """

    def apply_all_filters(self):
        """
        Apply all filters, select the brightest, and reset the source
        IDs.
        """
        cat = self.apply_filters()
        if cat is None:
            return None
        cat = cat.select_brightest()
        cat.reset_ids()
        return cat

    def to_table(self, *, columns=None):
        """
        Create a QTable of catalog properties.

        Parameters
        ----------
        columns : list of str, optional
            List of column names to include in the table. If `None`,
            uses ``self.default_columns``.

        Returns
        -------
        table : `~astropy.table.QTable`
            A table of the catalog properties.
        """
        # Replace with QTable in 4.0
        table = create_empty_deprecated_qtable(
            _DEPRECATED_ATTRIBUTES, since='3.0', until='4.0')

        table.meta.update(_get_meta())  # keep table.meta type
        if columns is None:
            if not self.default_columns:
                msg = ('default_columns attribute is not set; either '
                       'pass explicit column names or set '
                       'default_columns in the subclass __init__')
                raise AttributeError(msg)
            columns = self.default_columns
        for column in columns:
            table[column] = getattr(self, column)
        return table


class _StarFinderKernel:
    """
    Container class for a 2D Gaussian density enhancement kernel.

    The kernel has negative wings and sums to zero. It is used by both
    `DAOStarFinder` and `IRAFStarFinder`.

    Parameters
    ----------
    fwhm : float
        The full-width half-maximum (FWHM) of the major axis of the
        Gaussian kernel in units of pixels.

    ratio : float, optional
        The ratio of the minor and major axis standard deviations of
        the Gaussian kernel. ``ratio`` must be strictly positive and
        less than or equal to 1.0. The default is 1.0 (i.e., a circular
        Gaussian kernel).

    theta : float, optional
        The position angle (in degrees) of the major axis of the
        Gaussian kernel, measured counter-clockwise from the positive x
        axis.

    sigma_radius : float, optional
        The truncation radius of the Gaussian kernel in units
        of sigma (standard deviation) [``1 sigma = FWHM /
        (2.0 * sqrt(2.0 * log(2.0)))``]. The default is 1.5.

    normalize_zerosum : bool, optional
        Whether to normalize the Gaussian kernel to have zero sum. The
        default is `True`, which generates a density-enhancement kernel.

    Notes
    -----
    The class attributes include the dimensions of the elliptical kernel
    and the coefficients of a 2D elliptical Gaussian function expressed
    as:

    .. math::
        f(x, y) = A \\exp\\bigl(-g(x, y)\\bigr)

    where

    .. math::
        g(x, y) = a (x - x_0)^{2} + 2 b (x - x_0)(y - y_0)
                   + c (y - y_0)^{2}

    Attributes
    ----------
    data : 2D `~numpy.ndarray`
        The kernel data array, used for convolution.

    shape : tuple of int
        The ``(ny, nx)`` shape of ``data``.

    mask : 2D `~numpy.ndarray` of int
        Binary mask (1 inside the kernel footprint, 0 outside).
        Used for the peak-finding footprint, sharpness computation
        in `_DAOStarFinderCatalog`, and sky estimation in
        `_IRAFStarFinderCatalog`.

    rel_err : float
        The kernel relative error, used by `DAOStarFinder` to scale the
        detection threshold.

    gaussian_kernel_unmasked : 2D `~numpy.ndarray`
        The unmasked Gaussian kernel (peak normalized to 1), used by
        `_DAOStarFinderCatalog` for marginal fitting.

    x_sigma, y_sigma : float
        Standard deviations along the major and minor axes.

    x_radius, y_radius : int
        Half-widths of the kernel array in pixels.

    n_pixels : int
        Total number of pixels within the kernel ``mask``.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Gaussian_function
    """

    def __init__(self, fwhm, *, ratio=1.0, theta=0.0, sigma_radius=1.5,
                 normalize_zerosum=True):

        if np.ndim(fwhm) != 0:
            msg = 'fwhm must be a scalar value'
            raise TypeError(msg)

        if fwhm <= 0:
            msg = 'fwhm must be positive'
            raise ValueError(msg)

        if ratio <= 0 or ratio > 1:
            msg = 'ratio must be > 0 and <= 1.0'
            raise ValueError(msg)

        if sigma_radius <= 0:
            msg = 'sigma_radius must be positive'
            raise ValueError(msg)

        self.fwhm = fwhm
        self.ratio = ratio
        self.theta = theta % 360.0
        self.sigma_radius = sigma_radius
        self.x_sigma = self.fwhm * gaussian_fwhm_to_sigma
        self.y_sigma = self.x_sigma * self.ratio

        theta_radians = np.deg2rad(self.theta)
        cost = np.cos(theta_radians)
        sint = np.sin(theta_radians)
        x_sigma2 = self.x_sigma**2
        y_sigma2 = self.y_sigma**2

        a = (cost**2 / (2.0 * x_sigma2)) + (sint**2 / (2.0 * y_sigma2))
        # Counterclockwise rotation
        b = 0.5 * cost * sint * ((1.0 / x_sigma2) - (1.0 / y_sigma2))
        c = (sint**2 / (2.0 * x_sigma2)) + (cost**2 / (2.0 * y_sigma2))

        # Find the extent of an ellipse with radius = sigma_radius*sigma.
        # Solve for the horizontal and vertical tangents of an ellipse
        # defined by g(x,y) = f.
        f = self.sigma_radius**2 / 2.0
        denom = (a * c) - b**2

        # Ensure nx and ny are always odd.
        # The minimum kernel size is 5x5.
        nx = 2 * int(max(2, math.sqrt(c * f / denom))) + 1
        ny = 2 * int(max(2, math.sqrt(a * f / denom))) + 1

        self.x_radius = nx // 2
        self.y_radius = ny // 2

        # Define the kernel on a 2D grid
        xc = self.x_radius
        yc = self.y_radius
        yy, xx = np.mgrid[0:ny, 0:nx]
        circular_radius = np.sqrt((xx - xc)**2 + (yy - yc)**2)
        elliptical_radius = (a * (xx - xc)**2
                             + 2.0 * b * (xx - xc) * (yy - yc)
                             + c * (yy - yc)**2)

        self.mask = np.where(
            (elliptical_radius <= f)
            | (circular_radius <= 2.0), 1, 0).astype(int)
        self.n_pixels = self.mask.sum()

        # Central (peak) pixel of gaussian_kernel has a value of 1.0
        self.gaussian_kernel_unmasked = np.exp(-elliptical_radius)
        gaussian_kernel = self.gaussian_kernel_unmasked * self.mask

        # The denom represents (variance * n_pixels)
        denom = ((gaussian_kernel**2).sum()
                 - (gaussian_kernel.sum()**2 / self.n_pixels))
        self.rel_err = 1.0 / np.sqrt(denom)

        # Normalize the kernel to zero sum
        if normalize_zerosum:
            self.data = ((gaussian_kernel
                          - (gaussian_kernel.sum() / self.n_pixels))
                         / denom) * self.mask
        else:
            self.data = gaussian_kernel

        self.shape = self.data.shape

    def __repr__(self):
        params = ('fwhm', 'ratio', 'theta', 'sigma_radius')
        return make_repr(self, params)

    def __str__(self):
        params = ('fwhm', 'ratio', 'theta', 'sigma_radius')
        return make_repr(self, params, long=True)


def _validate_n_brightest(n_brightest):
    """
    Validate the ``n_brightest`` parameter.

    It must be >0 and an integer.

    Parameters
    ----------
    n_brightest : int, None, or bool
        The number of brightest sources to select. If `None`, all
        sources are selected. If a boolean is passed, a `TypeError` is
        raised.
    """
    if n_brightest is not None:
        if isinstance(n_brightest, bool):
            msg = 'n_brightest must be an integer'
            raise TypeError(msg)
        if n_brightest <= 0:
            msg = 'n_brightest must be > 0'
            raise ValueError(msg)
        bright_int = int(n_brightest)
        if bright_int != n_brightest:
            msg = 'n_brightest must be an integer'
            raise ValueError(msg)
        n_brightest = bright_int
    return n_brightest


def _handle_deprecated_range(old_lower, old_upper, new_range,
                             old_name, new_name, default_range):
    """
    Handle deprecated lower/upper bound parameters replaced by a single
    range parameter.

    Parameters
    ----------
    old_lower : float or `_DeprecatedDefault`
        The deprecated lower-bound parameter value.

    old_upper : float or `_DeprecatedDefault`
        The deprecated upper-bound parameter value.

    new_range : tuple of 2 floats or `None`
        The new range parameter value.

    old_name : str
        The base name of the deprecated parameters (e.g., ``'sharp'``
        for ``'sharplo'`` / ``'sharphi'``).

    new_name : str
        The name of the new range parameter (e.g.,
        ``'sharpness_range'``).

    default_range : tuple of 2 floats
        The default range values when ``new_range`` is `None`.

    Returns
    -------
    result : tuple of 2 floats or `None`
        The resolved range.
    """
    if old_lower is not _DEPR_DEFAULT or old_upper is not _DEPR_DEFAULT:
        msg = (f"The '{old_name}lo' and '{old_name}hi' parameters are "
               'deprecated and will be removed in a future version. '
               f"Use '{new_name}=(lower, upper)' instead.")
        warnings.warn(msg, AstropyDeprecationWarning)
        _default = new_range if new_range is not None else default_range
        lower = (old_lower if old_lower is not _DEPR_DEFAULT
                 else _default[0])
        upper = (old_upper if old_upper is not _DEPR_DEFAULT
                 else _default[1])
        return (lower, upper)
    return new_range


class _DeprecatedDefault:
    """
    Sentinel default value for a deprecated parameter.
    """

    def __repr__(self):
        return '<deprecated>'


_DEPR_DEFAULT = _DeprecatedDefault()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the rectangle module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import Angle, SkyCoord
from astropy.tests.helper import assert_quantity_allclose

from photutils.aperture.rectangle import (RectangularAnnulus,
                                          RectangularAperture,
                                          SkyRectangularAnnulus,
                                          SkyRectangularAperture)
from photutils.aperture.tests.test_aperture_common import BaseTestAperture
from photutils.utils._optional_deps import HAS_MATPLOTLIB

POSITIONS = [(10, 20), (30, 40), (50, 60), (70, 80)]
RA, DEC = np.transpose(POSITIONS)
SKYCOORD = SkyCoord(ra=RA, dec=DEC, unit='deg')
UNIT = u.arcsec
RADII = (0.0, -1.0, -np.inf)


class TestRectangularAperture(BaseTestAperture):
    aperture = RectangularAperture(POSITIONS, w=10.0, h=5.0, theta=np.pi / 2.0)

    @staticmethod
    @pytest.mark.parametrize('radius', RADII)
    def test_invalid_params(radius):
        match = "'w' must be a positive scalar"
        with pytest.raises(ValueError, match=match):
            RectangularAperture(POSITIONS, w=radius, h=5.0, theta=np.pi / 2.0)

        match = "'h' must be a positive scalar"
        with pytest.raises(ValueError, match=match):
            RectangularAperture(POSITIONS, w=10.0, h=radius, theta=np.pi / 2.0)

    def test_copy_eq(self):
        aper = self.aperture.copy()
        assert aper == self.aperture
        aper.w = 20.0
        assert aper != self.aperture

    def test_theta(self):
        assert isinstance(self.aperture.theta, u.Quantity)
        assert self.aperture.theta.unit == u.rad

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_to_patch_nonscalar(self):
        """
        Test that _to_patch returns a list for non-scalar apertures.
        """
        patches = self.aperture._to_patch()
        assert isinstance(patches, list)


class TestRectangularAnnulus(BaseTestAperture):
    aperture = RectangularAnnulus(POSITIONS, w_in=10.0, w_out=20.0, h_out=17,
                                  theta=np.pi / 3)

    @staticmethod
    @pytest.mark.parametrize('radius', RADII)
    def test_invalid_params(radius):
        match = "'w_in' must be a positive scalar"
        with pytest.raises(ValueError, match=match):
            RectangularAnnulus(POSITIONS, w_in=radius, w_out=20.0, h_out=17,
                               theta=np.pi / 3)

        match = "'w_out' must be greater than 'w_in'"
        with pytest.raises(ValueError, match=match):
            RectangularAnnulus(POSITIONS, w_in=10.0, w_out=radius, h_out=17,
                               theta=np.pi / 3)

        match = "'h_out' must be a positive scalar"
        with pytest.raises(ValueError, match=match):
            RectangularAnnulus(POSITIONS, w_in=10.0, w_out=20.0, h_out=radius,
                               theta=np.pi / 3)

        match = "'h_in' must be a positive scalar"
        with pytest.raises(ValueError, match=match):
            RectangularAnnulus(POSITIONS, w_in=10.0, w_out=20.0, h_out=17,
                               h_in=radius, theta=np.pi / 3)

    def test_copy_eq(self):
        aper = self.aperture.copy()
        assert aper == self.aperture
        aper.w_in = 2.0
        assert aper != self.aperture

    def test_theta(self):
        assert isinstance(self.aperture.theta, u.Quantity)
        assert self.aperture.theta.unit == u.rad

    def test_h_in_greater_than_h_out(self):
        """
        Test that a ValueError is raised when h_in >= h_out.
        """
        match = "'h_out' must be greater than 'h_in'"
        with pytest.raises(ValueError, match=match):
            RectangularAnnulus(POSITIONS, w_in=10.0, w_out=20.0, h_out=5.0,
                               h_in=8.0, theta=np.pi / 3)

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_to_patch_nonscalar(self):
        """
        Test that _to_patch returns a list for non-scalar apertures.
        """
        patches = self.aperture._to_patch()
        assert isinstance(patches, list)


class TestSkyRectangularAperture(BaseTestAperture):
    aperture = SkyRectangularAperture(SKYCOORD, w=10.0 * UNIT, h=5.0 * UNIT,
                                      theta=30 * u.deg)

    @staticmethod
    @pytest.mark.parametrize('radius', RADII)
    def test_invalid_params(radius):
        match = "'w' must be greater than zero"
        with pytest.raises(ValueError, match=match):
            SkyRectangularAperture(SKYCOORD, w=radius * UNIT, h=5.0 * UNIT,
                                   theta=30 * u.deg)

        match = "'h' must be greater than zero"
        with pytest.raises(ValueError, match=match):
            SkyRectangularAperture(SKYCOORD, w=10.0 * UNIT, h=radius * UNIT,
                                   theta=30 * u.deg)

    def test_copy_eq(self):
        aper = self.aperture.copy()
        assert aper == self.aperture
        aper.w = 20.0 * UNIT
        assert aper != self.aperture


class TestSkyRectangularAnnulus(BaseTestAperture):
    aperture = SkyRectangularAnnulus(SKYCOORD, w_in=10.0 * UNIT,
                                     w_out=20.0 * UNIT, h_out=17.0 * UNIT,
                                     theta=60 * u.deg)

    @staticmethod
    @pytest.mark.parametrize('radius', RADII)
    def test_invalid_params(radius):
        match = "'w_in' must be greater than zero"
        with pytest.raises(ValueError, match=match):
            SkyRectangularAnnulus(SKYCOORD, w_in=radius * UNIT,
                                  w_out=20.0 * UNIT, h_out=17.0 * UNIT,
                                  theta=60 * u.deg)

        match = "'w_out' must be greater than 'w_in'"
        with pytest.raises(ValueError, match=match):
            SkyRectangularAnnulus(SKYCOORD, w_in=10.0 * UNIT,
                                  w_out=radius * UNIT, h_out=17.0 * UNIT,
                                  theta=60 * u.deg)

        match = "'h_out' must be greater than zero"
        with pytest.raises(ValueError, match=match):
            SkyRectangularAnnulus(SKYCOORD, w_in=10.0 * UNIT,
                                  w_out=20.0 * UNIT, h_out=radius * UNIT,
                                  theta=60 * u.deg)

        match = "'h_in' must be greater than zero"
        with pytest.raises(ValueError, match=match):
            SkyRectangularAnnulus(SKYCOORD, w_in=10.0 * UNIT,
                                  w_out=20.0 * UNIT, h_out=17.0 * UNIT,
                                  h_in=radius * UNIT, theta=60 * u.deg)

    def test_copy_eq(self):
        aper = self.aperture.copy()
        assert aper == self.aperture
        aper.w_in = 2.0 * UNIT
        assert aper != self.aperture

    def test_h_in_greater_than_h_out(self):
        """
        Test that a ValueError is raised when h_in >= h_out.
        """
        match = "'h_out' must be greater than 'h_in'"
        with pytest.raises(ValueError, match=match):
            SkyRectangularAnnulus(SKYCOORD, w_in=10.0 * UNIT,
                                  w_out=20.0 * UNIT, h_out=5.0 * UNIT,
                                  h_in=8.0 * UNIT, theta=60 * u.deg)


def test_rectangle_theta_quantity():
    aper1 = RectangularAperture(POSITIONS, w=10.0, h=5.0, theta=np.pi / 2.0)
    theta = u.Quantity(90 * u.deg)
    aper2 = RectangularAperture(POSITIONS, w=10.0, h=5.0, theta=theta)
    theta = Angle(90 * u.deg)
    aper3 = RectangularAperture(POSITIONS, w=10.0, h=5.0, theta=theta)

    assert_quantity_allclose(aper1.theta, aper2.theta)
    assert_quantity_allclose(aper1.theta, aper3.theta)


def test_rectangle_annulus_theta_quantity():
    aper1 = RectangularAnnulus(POSITIONS, w_in=10.0, w_out=20.0, h_out=17,
                               theta=np.pi / 3)
    theta = u.Quantity(60 * u.deg)
    aper2 = RectangularAnnulus(POSITIONS, w_in=10.0, w_out=20.0, h_out=17,
                               theta=theta)
    theta = Angle(60 * u.deg)
    aper3 = RectangularAnnulus(POSITIONS, w_in=10.0, w_out=20.0, h_out=17,
                               theta=theta)

    assert_quantity_allclose(aper1.theta, aper2.theta)
    assert_quantity_allclose(aper1.theta, aper3.theta)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the _quantity_helpers module.
"""

import astropy.units as u
import numpy as np
import pytest
from numpy.testing import assert_equal

from photutils.utils._quantity_helpers import (check_units, isscalar,
                                               process_quantities)


@pytest.mark.parametrize('all_units', [False, True])
def test_units(all_units):
    """
    Test process_quantities with uniform units.
    """
    unit = u.Jy if all_units else 1.0
    arrs = (np.ones(3) * unit, np.ones(3) * unit, np.ones(3) * unit)
    names = ('a', 'b', 'c')

    arrs2, unit2 = process_quantities(arrs, names)
    if all_units:
        assert unit2 == unit
        for (arr, arr2) in zip(arrs, arrs2, strict=True):
            assert_equal(arr.value, arr2)
    else:
        assert unit2 is None
        assert arrs2 == arrs


def test_process_quantities_all_none():
    """
    Test that process_quantities with all None inputs returns None
    unit.
    """
    values, unit = process_quantities([None, None], ['a', 'b'])
    assert values == [None, None]
    assert unit is None


def test_isscalar():
    """
    Test isscalar with scalar and array inputs.
    """
    assert isscalar(1)
    assert isscalar(1.0 * u.m)
    assert not isscalar([1, 2, 3])
    assert not isscalar([1, 2, 3] * u.m)


def test_inputs():
    """
    Test that mismatched values and names lengths raises ValueError.
    """
    match = 'The number of values must match the number of names'
    with pytest.raises(ValueError, match=match):
        process_quantities([1, 2, 3], ['a', 'b'])
    with pytest.raises(ValueError, match=match):
        check_units([1, 2, 3], ['a', 'b'])


def test_check_units():
    """
    Test check_units for unit consistency checking.
    """
    # Valid: same units
    check_units((np.ones(3) * u.Jy, np.ones(3) * u.Jy), ('a', 'b'))

    # Valid: no units
    check_units((np.ones(3), np.ones(3)), ('a', 'b'))

    # Valid: with None values
    check_units((np.ones(3) * u.Jy, None), ('a', 'b'))


def test_mixed_units():
    """
    Test that check_units with mixed units raises ValueError.
    """
    arrs = (np.ones(3) * u.Jy, np.ones(3) * u.km)
    names = ('a', 'b')

    match = 'must all have the same units'
    with pytest.raises(ValueError, match=match):
        check_units(arrs, names)

    arrs = (np.ones(3) * u.Jy, np.ones(3))
    names = ('a', 'b')
    with pytest.raises(ValueError, match=match):
        check_units(arrs, names)

    unit = u.Jy
    arrs = (np.ones(3) * unit, np.ones(3), np.ones(3) * unit)
    names = ('a', 'b', 'c')
    with pytest.raises(ValueError, match=match):
        check_units(arrs, names)

    unit = u.Jy
    arrs = (np.ones(3) * unit, np.ones(3), np.ones(3) * u.km)
    names = ('a', 'b', 'c')
    with pytest.raises(ValueError, match=match):
        check_units(arrs, names)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Functional PSF models.
"""

import astropy.units as u
import numpy as np
from astropy.modeling import Fittable2DModel, Parameter
from astropy.modeling.utils import ellipse_extent
from astropy.units import UnitsError
from scipy.special import erf, j1, jn_zeros

__all__ = [
    'AiryDiskPSF',
    'CircularGaussianPRF',
    'CircularGaussianPSF',
    'CircularGaussianSigmaPRF',
    'GaussianPRF',
    'GaussianPSF',
    'MoffatPSF',
]

FLOAT_EPSILON = float(np.finfo(np.float32).tiny)
GAUSSIAN_FWHM_TO_SIGMA = 1.0 / (2.0 * np.sqrt(2.0 * np.log(2.0)))


def _gaussian_amplitude(flux, xsigma, ysigma):
    # output units should match the input flux units
    if isinstance(xsigma, u.Quantity):
        xsigma = xsigma.value
        ysigma = ysigma.value

    return flux / (2.0 * np.pi * xsigma * ysigma)


class GaussianPSF(Fittable2DModel):
    r"""
    A 2D Gaussian PSF model.

    This model is evaluated by sampling the 2D Gaussian at the input
    coordinates. The Gaussian is normalized such that the analytical
    integral over the entire 2D plane is equal to the total flux.

    Parameters
    ----------
    flux : float, optional
        Total integrated flux over the entire PSF.

    x_0 : float, optional
        Position of the peak along the x-axis.

    y_0 : float, optional
        Position of the peak along the y-axis.

    x_fwhm : float, optional
        The full width at half maximum (FWHM) of the Gaussian along the
        x axis.

    y_fwhm : float, optional
        FWHM of the Gaussian along the y axis.

    theta : float, optional
        The counterclockwise rotation angle either as a float (in
        degrees) or a `~astropy.units.Quantity` angle (optional).

    bbox_factor : float, optional
        The multiple of the x and y standard deviations (sigma) used to
        define the bounding box limits.

    **kwargs : dict, optional
        Additional optional keyword arguments to be passed to the
        `astropy.modeling.Model` base class.

    See Also
    --------
    CircularGaussianPSF, GaussianPRF, CircularGaussianPRF, MoffatPSF

    Notes
    -----
    The Gaussian function is defined as:

    .. math::

        f(x, y) = \frac{F}{2 \pi \sigma_{x} \sigma_{y}}
                  \exp \left( -a\left(x - x_{0}\right)^{2}
                  - b \left(x - x_{0}\right) \left(y - y_{0}\right)
                  - c \left(y - y_{0}\right)^{2} \right)

    where :math:`F` denotes the total integrated flux, :math:`(x_{0},
    y_{0})` denotes the position of the peak, and :math:`\sigma_{x}` and
    :math:`\sigma_{y}` denote are the standard deviations along the x
    and y axes, respectively.

    .. math::

        a = \frac{\cos^{2}{\theta}}{2 \sigma_{x}^{2}}
            + \frac{\sin^{2}{\theta}}{2 \sigma_{y}^{2}}

        b = \frac{\sin{2 \theta}}{2 \sigma_{x}^{2}}
            - \frac{\sin{2 \theta}}{2 \sigma_{y}^{2}}

        c = \frac{\sin^{2}{\theta}} {2 \sigma_{x}^{2}}
            + \frac{\cos^{2}{\theta}}{2 \sigma_{y}^{2}}

    where :math:`\theta` is the rotation angle of the Gaussian.

    The FWHMs of the Gaussian along the x and y axes are given by:

    .. math::

        \rm{FWHM}_{x} = 2 \sigma_{x} \sqrt{2 \ln{2}}

        \rm{FWHM}_{y} = 2 \sigma_{y} \sqrt{2 \ln{2}}

    The model is normalized such that:

    .. math::

        \int_{-\infty}^{\infty} \int_{-\infty}^{\infty} f(x, y) \,dx \,dy = F

    The ``x_fwhm``, ``y_fwhm``, and ``theta`` parameters are fixed by
    default. If you wish to fit these parameters, set the ``fixed``
    attribute to `False`, e.g.,::

        >>> from photutils.psf import GaussianPSF
        >>> model = GaussianPSF()
        >>> model.x_fwhm.fixed = False
        >>> model.y_fwhm.fixed = False
        >>> model.theta.fixed = False

    By default, the ``x_fwhm`` and ``y_fwhm`` parameters are bounded to
    be strictly positive.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Gaussian_function

    Examples
    --------
    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        import numpy as np
        from photutils.psf import GaussianPSF

        model = GaussianPSF(flux=71.4, x_0=24.3, y_0=25.2, x_fwhm=10.1,
                            y_fwhm=5.82, theta=21.7)
        yy, xx = np.mgrid[0:51, 0:51]
        data = model(xx, yy)
        fig, ax = plt.subplots()
        ax.imshow(data, origin='lower')
    """

    flux = Parameter(
        default=1, description='Total integrated flux over the entire PSF.')
    x_0 = Parameter(
        default=0, description='Position of the peak along the x axis')
    y_0 = Parameter(
        default=0, description='Position of the peak along the y axis')
    x_fwhm = Parameter(
        default=1,
        bounds=(FLOAT_EPSILON, None),
        fixed=True,
        description='FWHM of the Gaussian along the x axis')
    y_fwhm = Parameter(
        default=1,
        bounds=(FLOAT_EPSILON, None),
        fixed=True,
        description='FWHM of the Gaussian along the y axis')
    theta = Parameter(
        default=0.0, description=('CCW rotation angle either as a float (in '
                                  'degrees) or a Quantity angle (optional)'),
        fixed=True)

    def __init__(self, *, flux=flux.default, x_0=x_0.default, y_0=y_0.default,
                 x_fwhm=x_fwhm.default, y_fwhm=y_fwhm.default,
                 theta=theta.default, bbox_factor=5.5, **kwargs):
        super().__init__(flux=flux, x_0=x_0, y_0=y_0, x_fwhm=x_fwhm,
                         y_fwhm=y_fwhm, theta=theta, **kwargs)
        self.bbox_factor = bbox_factor

    @property
    def amplitude(self):
        """
        The peak amplitude of the Gaussian.
        """
        return _gaussian_amplitude(self.flux, self.x_sigma, self.y_sigma)

    @property
    def x_sigma(self):
        """
        Gaussian sigma (standard deviation) along the x-axis.
        """
        return self.x_fwhm * GAUSSIAN_FWHM_TO_SIGMA

    @property
    def y_sigma(self):
        """
        Gaussian sigma (standard deviation) along the y-axis.
        """
        return self.y_fwhm * GAUSSIAN_FWHM_TO_SIGMA

    def _calc_bounding_box(self, *, factor=5.5):
        """
        Calculate a bounding box defining the limits of the model.

        The limits are adjusted for rotation.

        Parameters
        ----------
        factor : float, optional
            The multiple of the x and y standard deviations (sigma) used
            to define the limits.

        Returns
        -------
        bbox : tuple
            A bounding box defining the ((y_min, y_max), (x_min, x_max))
            limits of the model.
        """
        a = factor * self.x_sigma
        b = factor * self.y_sigma
        dx, dy = ellipse_extent(a, b, self.theta)
        return ((self.y_0 - dy, self.y_0 + dy), (self.x_0 - dx, self.x_0 + dx))

    @property
    def bounding_box(self):
        """
        The bounding box of the model.

        Examples
        --------
        >>> from photutils.psf import GaussianPSF
        >>> model = GaussianPSF(x_0=0, y_0=0, x_fwhm=2, y_fwhm=3)
        >>> model.bounding_box  # doctest: +FLOAT_CMP
        ModelBoundingBox(
            intervals={
                x: Interval(lower=-4.671269901584105, upper=4.671269901584105)
                y: Interval(lower=-7.006904852376157, upper=7.006904852376157)
            }
            model=GaussianPSF(inputs=('x', 'y'))
            order='C'
        )
        >>> model.bbox_factor = 7
        >>> model.bounding_box  # doctest: +FLOAT_CMP
        ModelBoundingBox(
            intervals={
                x: Interval(lower=-5.945252602016134, upper=5.945252602016134)
                y: Interval(lower=-8.9178789030242, upper=8.9178789030242)
            }
            model=GaussianPSF(inputs=('x', 'y'))
            order='C'
        )
        """
        return self._calc_bounding_box(factor=self.bbox_factor)

    def evaluate(self, x, y, flux, x_0, y_0, x_fwhm, y_fwhm, theta):
        """
        Calculate the value of the 2D Gaussian model at the input
        coordinates for the given model parameters.

        Parameters
        ----------
        x, y : float or array_like
            The x and y coordinates at which to evaluate the model.

        flux : float
            Total integrated flux over the entire PSF.

        x_0, y_0 : float
            Position of the peak along the x and y axes.

        x_fwhm, y_fwhm : float
            FWHM of the Gaussian along the x and y axes.

        theta : float
            The counterclockwise rotation angle either as a float (in
            degrees) or a `~astropy.units.Quantity` angle (optional).

        Returns
        -------
        result : `~numpy.ndarray`
            The value of the model evaluated at the input coordinates.
        """
        if not isinstance(theta, u.Quantity):
            theta = np.deg2rad(theta)
        cost2 = np.cos(theta) ** 2
        sint2 = np.sin(theta) ** 2
        sin2t = np.sin(2.0 * theta)
        xstd = x_fwhm * GAUSSIAN_FWHM_TO_SIGMA
        ystd = y_fwhm * GAUSSIAN_FWHM_TO_SIGMA
        xstd2 = xstd ** 2
        ystd2 = ystd ** 2
        xdiff = x - x_0
        ydiff = y - y_0
        a = 0.5 * ((cost2 / xstd2) + (sint2 / ystd2))
        b = 0.5 * ((sin2t / xstd2) - (sin2t / ystd2))
        c = 0.5 * ((sint2 / xstd2) + (cost2 / ystd2))

        # output units should match the input flux units
        if isinstance(xstd, u.Quantity):
            xstd = xstd.value
            ystd = ystd.value

        amplitude = flux / (2 * np.pi * xstd * ystd)
        return amplitude * np.exp(
            -(a * xdiff**2) - (b * xdiff * ydiff) - (c * ydiff**2))

    @staticmethod
    def fit_deriv(x, y, flux, x_0, y_0, x_fwhm, y_fwhm, theta):
        """
        Calculate the partial derivatives of the 2D Gaussian function
        with respect to the parameters.

        Parameters
        ----------
        x, y : float or array_like
            The x and y coordinates at which to evaluate the model.

        flux : float
            Total integrated flux over the entire PSF.

        x_0, y_0 : float
            Position of the peak along the x and y axes.

        x_fwhm, y_fwhm : float
            FWHM of the Gaussian along the x and y axes.

        theta : float
            The counterclockwise rotation angle either as a float (in
            degrees) or a `~astropy.units.Quantity` angle (optional).

        Returns
        -------
        result : list of `~numpy.ndarray`
            The list of partial derivatives with respect to each
            parameter.
        """
        if not isinstance(theta, u.Quantity):
            theta = np.deg2rad(theta)

        cost = np.cos(theta)
        sint = np.sin(theta)
        cost2 = cost ** 2
        sint2 = sint ** 2
        cos2t = np.cos(2.0 * theta)
        sin2t = np.sin(2.0 * theta)
        xstd = x_fwhm * GAUSSIAN_FWHM_TO_SIGMA
        ystd = y_fwhm * GAUSSIAN_FWHM_TO_SIGMA
        xstd2 = xstd ** 2
        ystd2 = ystd ** 2
        xstd3 = xstd ** 3
        ystd3 = ystd ** 3
        xdiff = x - x_0
        ydiff = y - y_0
        xdiff2 = xdiff ** 2
        ydiff2 = ydiff ** 2
        a = 0.5 * ((cost2 / xstd2) + (sint2 / ystd2))
        b = 0.5 * ((sin2t / xstd2) - (sin2t / ystd2))
        c = 0.5 * ((sint2 / xstd2) + (cost2 / ystd2))

        amplitude = flux / (2 * np.pi * xstd * ystd)
        exp = np.exp(-(a * xdiff2) - (b * xdiff * ydiff) - (c * ydiff2))
        g = amplitude * exp

        da_dtheta = sint * cost * ((1.0 / ystd2) - (1.0 / xstd2))
        db_dtheta = (cos2t / xstd2) - (cos2t / ystd2)
        dc_dtheta = -da_dtheta

        da_dxstd = -cost2 / xstd3
        db_dxstd = -sin2t / xstd3
        dc_dxstd = -sint2 / xstd3

        da_dystd = -sint2 / ystd3
        db_dystd = sin2t / ystd3
        dc_dystd = -cost2 / ystd3

        dg_dflux = g / flux
        dg_dx_0 = g * ((2.0 * a * xdiff) + (b * ydiff))
        dg_dy_0 = g * ((b * xdiff) + (2.0 * c * ydiff))

        damp_dxstd = -amplitude / xstd
        damp_dystd = -amplitude / ystd
        dexp_dxstd = -exp * (da_dxstd * xdiff2
                             + db_dxstd * xdiff * ydiff
                             + dc_dxstd * ydiff2)
        dexp_dystd = -exp * (da_dystd * xdiff2
                             + db_dystd * xdiff * ydiff
                             + dc_dystd * ydiff2)
        dg_dxstd = damp_dxstd * exp + amplitude * dexp_dxstd
        dg_dystd = damp_dystd * exp + amplitude * dexp_dystd

        # chain rule for change of variables from sigma to fwhm
        # std => fwhm * GAUSSIAN_FWHM_TO_SIGMA
        # dstd/dfwhm => GAUSSIAN_FWHM_TO_SIGMA
        dg_dxfwhm = dg_dxstd * GAUSSIAN_FWHM_TO_SIGMA
        dg_dyfwhm = dg_dystd * GAUSSIAN_FWHM_TO_SIGMA

        dg_dtheta = g * (-(da_dtheta * xdiff2 + db_dtheta * xdiff * ydiff
                           + dc_dtheta * ydiff2))
        # chain rule for unit change;
        # theta[rad] => theta[deg] * pi / 180; drad/dtheta = pi / 180
        dg_dtheta *= np.pi / 180.0

        return [dg_dflux, dg_dx_0, dg_dy_0, dg_dxfwhm, dg_dyfwhm, dg_dtheta]

    @property
    def input_units(self):
        """
        The input units of the model.
        """
        x_unit = self.x_0.input_unit
        y_unit = self.y_0.input_unit
        if x_unit is None and y_unit is None:
            return None

        return {self.inputs[0]: x_unit, self.inputs[1]: y_unit}

    def _parameter_units_for_data_units(self, inputs_unit, outputs_unit):
        # Note that here we need to make sure that x and y are in the same
        # units otherwise this can lead to issues since rotation is not well
        # defined.
        if inputs_unit[self.inputs[0]] != inputs_unit[self.inputs[1]]:
            msg = "Units of 'x' and 'y' inputs should match"
            raise UnitsError(msg)

        return {'x_0': inputs_unit[self.inputs[0]],
                'y_0': inputs_unit[self.inputs[0]],
                'x_fwhm': inputs_unit[self.inputs[0]],
                'y_fwhm': inputs_unit[self.inputs[0]],
                'theta': u.deg,
                'flux': outputs_unit[self.outputs[0]]}


class CircularGaussianPSF(Fittable2DModel):
    r"""
    A circular 2D Gaussian PSF model.

    This model is evaluated by sampling the 2D Gaussian at the input
    coordinates. The Gaussian is normalized such that the analytical
    integral over the entire 2D plane is equal to the total flux.

    Parameters
    ----------
    flux : float, optional
        Total integrated flux over the entire PSF.

    x_0 : float, optional
        Position of the peak along the x-axis.

    y_0 : float, optional
        Position of the peak along the y-axis.

    fwhm : float, optional
        The full width at half maximum (FWHM) of the Gaussian.

    bbox_factor : float, optional
        The multiple of the standard deviation (sigma) used to define
        the bounding box limits.

    **kwargs : dict, optional
        Additional optional keyword arguments to be passed to the
        `astropy.modeling.Model` base class.

    See Also
    --------
    GaussianPSF, GaussianPRF, CircularGaussianPRF, MoffatPSF

    Notes
    -----
    The circular Gaussian function is defined as:

    .. math::

        f(x, y) = \frac{F}{2 \pi \sigma^{2}}
                  \exp \left( {\frac{-(x - x_{0})^{2} - (y - y_{0})^{2}}
                             {2 \sigma^{2}}} \right)

    where :math:`F` is the total integrated flux, :math:`(x_{0}, y_{0})`
    is the position of the peak, and :math:`\sigma` is the standard
    deviation, respectively.

    The FWHM of the Gaussian is given by:

    .. math::

        \rm{FWHM} = 2 \sigma \sqrt{2 \ln{2}}

    The model is normalized such that:

    .. math::

        \int_{-\infty}^{\infty} \int_{-\infty}^{\infty} f(x, y) \,dx \,dy = F

    The ``fwhm`` parameter is fixed by default. If you wish to fit this
    parameter, set the ``fixed`` attribute to `False`, e.g.,::

        >>> from photutils.psf import CircularGaussianPSF
        >>> model = CircularGaussianPSF()
        >>> model.fwhm.fixed = False

    By default, the ``fwhm`` parameter is bounded to be strictly
    positive.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Gaussian_function

    Examples
    --------
    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        import numpy as np
        from photutils.psf import CircularGaussianPSF

        model = CircularGaussianPSF(flux=71.4, x_0=24.3, y_0=25.2, fwhm=10.1)
        yy, xx = np.mgrid[0:51, 0:51]
        data = model(xx, yy)
        fig, ax = plt.subplots()
        ax.imshow(data, origin='lower')
    """

    flux = Parameter(
        default=1, description='Total integrated flux over the entire PSF.')
    x_0 = Parameter(
        default=0, description='Position of the peak along the x axis')
    y_0 = Parameter(
        default=0, description='Position of the peak along the y axis')
    fwhm = Parameter(
        default=1,
        bounds=(FLOAT_EPSILON, None),
        fixed=True,
        description='FWHM of the Gaussian')

    def __init__(self, *, flux=flux.default, x_0=x_0.default, y_0=y_0.default,
                 fwhm=fwhm.default, bbox_factor=5.5, **kwargs):
        super().__init__(flux=flux, x_0=x_0, y_0=y_0, fwhm=fwhm, **kwargs)
        self.bbox_factor = bbox_factor

    @property
    def amplitude(self):
        """
        The peak amplitude of the Gaussian.
        """
        return _gaussian_amplitude(self.flux, self.sigma, self.sigma)

    @property
    def sigma(self):
        """
        Gaussian sigma (standard deviation).
        """
        return self.fwhm * GAUSSIAN_FWHM_TO_SIGMA

    def _calc_bounding_box(self, *, factor=5.5):
        """
        Calculate a bounding box defining the limits of the model.

        Parameters
        ----------
        factor : float, optional
            The multiple of the standard deviations (sigma) used to
            define the limits.

        Returns
        -------
        bbox : tuple
            A bounding box defining the ((y_min, y_max), (x_min, x_max))
            limits of the model.
        """
        delta = factor * self.sigma
        return ((self.y_0 - delta, self.y_0 + delta),
                (self.x_0 - delta, self.x_0 + delta))

    @property
    def bounding_box(self):
        """
        The bounding box of the model.

        Examples
        --------
        >>> from photutils.psf import CircularGaussianPSF
        >>> model = CircularGaussianPSF(x_0=0, y_0=0, fwhm=2)
        >>> model.bounding_box  # doctest: +FLOAT_CMP
        ModelBoundingBox(
            intervals={
                x: Interval(lower=-4.671269901584105, upper=4.671269901584105)
                y: Interval(lower=-4.671269901584105, upper=4.671269901584105)
            }
            model=CircularGaussianPSF(inputs=('x', 'y'))
            order='C'
        )
        >>> model.bbox_factor = 7
        >>> model.bounding_box  # doctest: +FLOAT_CMP
        ModelBoundingBox(
            intervals={
                x: Interval(lower=-5.945252602016134, upper=5.945252602016134)
                y: Interval(lower=-5.945252602016134, upper=5.945252602016134)
            }
            model=CircularGaussianPSF(inputs=('x', 'y'))
            order='C'
        )
        """
        return self._calc_bounding_box(factor=self.bbox_factor)

    def evaluate(self, x, y, flux, x_0, y_0, fwhm):
        """
        Calculate the value of the 2D Gaussian model at the input
        coordinates for the given model parameters.

        Parameters
        ----------
        x, y : float or array_like
            The x and y coordinates at which to evaluate the model.

        flux : float
            Total integrated flux over the entire PSF.

        x_0, y_0 : float
            Position of the peak along the x and y axes.

        fwhm : float
            FWHM of the Gaussian.

        Returns
        -------
        result : `~numpy.ndarray`
            The value of the model evaluated at the input coordinates.
        """
        sigma2 = (fwhm * GAUSSIAN_FWHM_TO_SIGMA) ** 2

        # output units should match the input flux units
        sigma2_norm = sigma2
        if isinstance(sigma2, u.Quantity):
            sigma2_norm = sigma2.value

        amplitude = flux / (2 * np.pi * sigma2_norm)
        return amplitude * np.exp(-0.5 * ((x - x_0) ** 2 + (y - y_0) ** 2)
                                  / sigma2)

    @staticmethod
    def fit_deriv(x, y, flux, x_0, y_0, fwhm):
        """
        Calculate the partial derivatives of the 2D Gaussian function
        with respect to the parameters.

        Parameters
        ----------
        x, y : float or array_like
            The x and y coordinates at which to evaluate the model.

        flux : float
            Total integrated flux over the entire PSF.

        x_0, y_0 : float
            Position of the peak along the x and y axes.

        fwhm : float
            FWHM of the Gaussian.

        Returns
        -------
        result : list of `~numpy.ndarray`
            The list of partial derivatives with respect to each
            parameter.
        """
        return GaussianPSF().fit_deriv(x, y, flux, x_0, y_0, fwhm, fwhm,
                                       0.0)[:-2]

    @property
    def input_units(self):
        """
        The input units of the model.
        """
        x_unit = self.x_0.input_unit
        y_unit = self.y_0.input_unit
        if x_unit is None and y_unit is None:
            return None

        return {self.inputs[0]: x_unit, self.inputs[1]: y_unit}

    def _parameter_units_for_data_units(self, inputs_unit, outputs_unit):
        return {'x_0': inputs_unit[self.inputs[0]],
                'y_0': inputs_unit[self.inputs[0]],
                'fwhm': inputs_unit[self.inputs[0]],
                'flux': outputs_unit[self.outputs[0]]}


class GaussianPRF(Fittable2DModel):
    r"""
    A 2D Gaussian PSF model integrated over pixels.

    This model is evaluated by integrating the 2D Gaussian over the
    input coordinate pixels, and is equivalent to assuming the PSF is
    2D Gaussian at a *sub-pixel* level. Because it is integrated over
    pixels, this model is considered a PRF instead of a PSF.

    The Gaussian is normalized such that the analytical integral over
    the entire 2D plane is equal to the total flux.

    Parameters
    ----------
    flux : float, optional
        Total integrated flux over the entire PSF.

    x_0 : float, optional
        Position of the peak along the x-axis.

    y_0 : float, optional
        Position of the peak along the y-axis.

    x_fwhm : float, optional
        The full width at half maximum (FWHM) of the Gaussian along the
        x axis.

    y_fwhm : float, optional
        FWHM of the Gaussian along the y axis.

    theta : float, optional
        The counterclockwise rotation angle either as a float (in
        degrees) or a `~astropy.units.Quantity` angle (optional).

    bbox_factor : float, optional
        The multiple of the x and y standard deviations (sigma) used to
        define the bounding_box limits.

    **kwargs : dict, optional
        Additional optional keyword arguments to be passed to the
        `astropy.modeling.Model` base class.

    See Also
    --------
    GaussianPSF, CircularGaussianPSF, CircularGaussianPRF, MoffatPSF

    Notes
    -----
    The Gaussian function is defined as:

    .. math::

        f(x, y) =
            \frac{F}{4}
            \left[
                {\rm erf} \left(
                    \frac{x^\prime + 0.5}{\sqrt{2} \sigma_{x}} \right) -
                {\rm erf} \left(
                    \frac{x^\prime - 0.5}{\sqrt{2} \sigma_{x}} \right)
            \right]
            \left[
                {\rm erf} \left(
                    \frac{y^\prime + 0.5}{\sqrt{2} \sigma_{y}} \right) -
                {\rm erf} \left(
                    \frac{y^\prime - 0.5}{\sqrt{2} \sigma_{y}} \right)
            \right]

    where :math:`F` is the total integrated flux, :math:`\sigma_{x}`
    and :math:`\sigma_{y}` are the standard deviations along the x
    and y axes, respectively, and :math:`{\rm erf}` denotes the error
    function.

    .. math::

        x^\prime = (x - x_0) \cos(\theta) + (y - y_0) \sin(\theta)

        y^\prime = -(x - x_0) \sin(\theta) + (y - y_0) \cos(\theta)

    where :math:`(x_{0}, y_{0})` is the position of the peak and
    :math:`\theta` is the rotation angle of the Gaussian.

    The FWHMs of the Gaussian along the x and y axes are given by:

    .. math::

        \rm{FWHM}_{x} = 2 \sigma_{x} \sqrt{2 \ln{2}}

        \rm{FWHM}_{y} = 2 \sigma_{y} \sqrt{2 \ln{2}}

    The model is normalized such that:

    .. math::

        \int_{-\infty}^{\infty} \int_{-\infty}^{\infty} f(x, y) \,dx \,dy = F

    The ``x_fwhm``, ``y_fwhm``, and ``theta`` parameters are fixed by
    default. If you wish to fit these parameters, set the ``fixed``
    attribute to `False`, e.g.,::

        >>> from photutils.psf import GaussianPRF
        >>> model = GaussianPRF()
        >>> model.x_fwhm.fixed = False
        >>> model.y_fwhm.fixed = False
        >>> model.theta.fixed = False

    By default, the ``x_fwhm`` and ``y_fwhm`` parameters are bounded to
    be strictly positive.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Gaussian_function

    Examples
    --------
    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        import numpy as np
        from photutils.psf import GaussianPRF

        model = GaussianPRF(flux=71.4, x_0=24.3, y_0=25.2, x_fwhm=10.1,
                            y_fwhm=5.82, theta=21.7)
        yy, xx = np.mgrid[0:51, 0:51]
        data = model(xx, yy)
        fig, ax = plt.subplots()
        ax.imshow(data, origin='lower')
    """

    flux = Parameter(
        default=1, description='Total integrated flux over the entire PSF.')
    x_0 = Parameter(
        default=0, description='Position of the peak along the x axis')
    y_0 = Parameter(
        default=0, description='Position of the peak along the y axis')
    x_fwhm = Parameter(
        default=1,
        bounds=(FLOAT_EPSILON, None),
        fixed=True,
        description='FWHM of the Gaussian along the x axis')
    y_fwhm = Parameter(
        default=1,
        bounds=(FLOAT_EPSILON, None),
        fixed=True,
        description='FWHM of the Gaussian along the y axis')
    theta = Parameter(
        default=0.0, description=('CCW rotation angle either as a float (in '
                                  'degrees) or a Quantity angle (optional)'),
        fixed=True)

    def __init__(self, *, flux=flux.default, x_0=x_0.default, y_0=y_0.default,
                 x_fwhm=x_fwhm.default, y_fwhm=y_fwhm.default,
                 theta=theta.default, bbox_factor=5.5, **kwargs):
        super().__init__(flux=flux, x_0=x_0, y_0=y_0, x_fwhm=x_fwhm,
                         y_fwhm=y_fwhm, theta=theta, **kwargs)
        self.bbox_factor = bbox_factor

    @property
    def amplitude(self):
        """
        The peak amplitude of the Gaussian.
        """
        return _gaussian_amplitude(self.flux, self.x_sigma, self.y_sigma)

    @property
    def x_sigma(self):
        """
        Gaussian sigma (standard deviation) along the x-axis.
        """
        return self.x_fwhm * GAUSSIAN_FWHM_TO_SIGMA

    @property
    def y_sigma(self):
        """
        Gaussian sigma (standard deviation) along the y-axis.
        """
        return self.y_fwhm * GAUSSIAN_FWHM_TO_SIGMA

    def _calc_bounding_box(self, *, factor=5.5):
        """
        Calculate a bounding box defining the limits of the model.

        The limits are adjusted for rotation.

        Parameters
        ----------
        factor : float, optional
            The multiple of the x and y FWHMs used to define the limits.
            zzzz

        Returns
        -------
        bbox : tuple
            A bounding box defining the ((y_min, y_max), (x_min, x_max))
            limits of the model.
        """
        a = factor * self.x_sigma
        b = factor * self.y_sigma
        dx, dy = ellipse_extent(a, b, self.theta)
        return ((self.y_0 - dy, self.y_0 + dy), (self.x_0 - dx, self.x_0 + dx))

    @property
    def bounding_box(self):
        """
        The bounding box of the model.

        Examples
        --------
        >>> from photutils.psf import GaussianPRF
        >>> model = GaussianPRF(x_0=0, y_0=0, x_fwhm=2, y_fwhm=3)
        >>> model.bounding_box  # doctest: +FLOAT_CMP
        ModelBoundingBox(
            intervals={
                x: Interval(lower=-4.671269901584105, upper=4.671269901584105)
                y: Interval(lower=-7.006904852376157, upper=7.006904852376157)
            }
            model=GaussianPRF(inputs=('x', 'y'))
            order='C'
        )
        >>> model.bbox_factor = 7
        >>> model.bounding_box  # doctest: +FLOAT_CMP
        ModelBoundingBox(
            intervals={
                x: Interval(lower=-5.945252602016134, upper=5.945252602016134)
                y: Interval(lower=-8.9178789030242, upper=8.9178789030242)
            }
            model=GaussianPRF(inputs=('x', 'y'))
            order='C'
        )
        """
        return self._calc_bounding_box(factor=self.bbox_factor)

    def evaluate(self, x, y, flux, x_0, y_0, x_fwhm, y_fwhm, theta):
        """
        Calculate the value of the 2D Gaussian model at the input
        coordinates for the given model parameters.

        Parameters
        ----------
        x, y : float or array_like
            The x and y coordinates at which to evaluate the model.

        flux : float
            Total integrated flux over the entire PSF.

        x_0, y_0 : float
            Position of the peak along the x and y axes.

        x_fwhm, y_fwhm : float
            FWHM of the Gaussian along the x and y axes.

        theta : float
            The counterclockwise rotation angle either as a float (in
            degrees) or a `~astropy.units.Quantity` angle (optional).

        Returns
        -------
        result : `~numpy.ndarray`
            The value of the model evaluated at the input coordinates.
        """
        if not isinstance(theta, u.Quantity):
            theta = np.deg2rad(theta)

        x_sigma = x_fwhm * GAUSSIAN_FWHM_TO_SIGMA
        y_sigma = y_fwhm * GAUSSIAN_FWHM_TO_SIGMA
        dx = x - x_0
        dy = y - y_0
        cost = np.cos(theta)
        sint = np.sin(theta)
        x0 = dx * cost + dy * sint
        y0 = -dx * sint + dy * cost

        dpix = 0.5
        if isinstance(x0, u.Quantity):
            dpix <<= x0.unit

        return (flux / 4.0
                * ((erf((x0 + dpix) / (np.sqrt(2) * x_sigma))
                    - erf((x0 - dpix) / (np.sqrt(2) * x_sigma)))
                   * (erf((y0 + dpix) / (np.sqrt(2) * y_sigma))
                      - erf((y0 - dpix) / (np.sqrt(2) * y_sigma)))))

    @property
    def input_units(self):
        """
        The input units of the model.
        """
        x_unit = self.x_0.input_unit
        y_unit = self.y_0.input_unit
        if x_unit is None and y_unit is None:
            return None

        return {self.inputs[0]: x_unit, self.inputs[1]: y_unit}

    def _parameter_units_for_data_units(self, inputs_unit, outputs_unit):
        # Note that here we need to make sure that x and y are in the same
        # units otherwise this can lead to issues since rotation is not well
        # defined.
        if inputs_unit[self.inputs[0]] != inputs_unit[self.inputs[1]]:
            msg = "Units of 'x' and 'y' inputs should match"
            raise UnitsError(msg)

        return {'x_0': inputs_unit[self.inputs[0]],
                'y_0': inputs_unit[self.inputs[0]],
                'x_fwhm': inputs_unit[self.inputs[0]],
                'y_fwhm': inputs_unit[self.inputs[0]],
                'theta': u.deg,
                'flux': outputs_unit[self.outputs[0]]}


class CircularGaussianPRF(Fittable2DModel):
    r"""
    A circular 2D Gaussian PSF model integrated over pixels.

    This model is evaluated by integrating the 2D Gaussian over the
    input coordinate pixels, and is equivalent to assuming the PSF is
    2D Gaussian at a *sub-pixel* level. Because it is integrated over
    pixels, this model is considered a PRF instead of a PSF.

    The Gaussian is normalized such that the analytical integral over
    the entire 2D plane is equal to the total flux.

    Parameters
    ----------
    flux : float, optional
        Total integrated flux over the entire PSF.

    x_0 : float, optional
        Position of the peak along the x-axis.

    y_0 : float, optional
        Position of the peak along the y-axis.

    fwhm : float, optional
        The full width at half maximum (FWHM) of the Gaussian.

    bbox_factor : float, optional
        The multiple of the standard deviation (sigma) used to define
        the bounding box limits.

    **kwargs : dict, optional
        Additional optional keyword arguments to be passed to the
        `astropy.modeling.Model` base class.

    See Also
    --------
    GaussianPRF, GaussianPSF, CircularGaussianPSF, MoffatPSF

    Notes
    -----
    The circular Gaussian function is defined as:

    .. math::

        f(x, y) =
            \frac{F}{4}
            \left[
                {\rm erf} \left(
                    \frac{x - x_0 + 0.5}{\sqrt{2} \sigma} \right) -
                {\rm erf} \left(
                    \frac{x - x_0 - 0.5}{\sqrt{2} \sigma} \right)
            \right]
            \left[
                {\rm erf} \left(
                    \frac{y - y_0 + 0.5}{\sqrt{2} \sigma} \right) -
                {\rm erf} \left(
                    \frac{y - y_0 - 0.5}{\sqrt{2} \sigma} \right)
            \right]

    where :math:`F` is the total integrated flux, :math:`(x_{0},
    y_{0})` is the position of the peak, :math:`\sigma` is the standard
    deviation of the Gaussian, and :math:`{\rm erf}` denotes the error
    function.

    The FWHM of the Gaussian is given by:

    .. math::

        \rm{FWHM} = 2 \sigma \sqrt{2 \ln{2}}

    The model is normalized such that:

    .. math::

        \int_{-\infty}^{\infty} \int_{-\infty}^{\infty} f(x, y) \,dx \,dy = F

    The ``fwhm`` parameter is fixed by default. If you wish to fit this
    parameter, set the ``fixed`` attribute to `False`, e.g.,::

        >>> from photutils.psf import CircularGaussianPRF
        >>> model = CircularGaussianPRF()
        >>> model.fwhm.fixed = False

    By default, the ``fwhm`` parameter is bounded to be strictly
    positive.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Gaussian_function

    Examples
    --------
    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        import numpy as np
        from photutils.psf import CircularGaussianPRF

        model = CircularGaussianPRF(flux=71.4, x_0=24.3, y_0=25.2, fwhm=10.1)
        yy, xx = np.mgrid[0:51, 0:51]
        data = model(xx, yy)
        fig, ax = plt.subplots()
        ax.imshow(data, origin='lower')
    """

    flux = Parameter(
        default=1, description='Total integrated flux over the entire PSF.')
    x_0 = Parameter(
        default=0, description='Position of the peak along the x axis')
    y_0 = Parameter(
        default=0, description='Position of the peak along the y axis')
    fwhm = Parameter(
        default=1,
        bounds=(FLOAT_EPSILON, None),
        fixed=True,
        description='FWHM of the Gaussian')

    def __init__(self, *, flux=flux.default, x_0=x_0.default, y_0=y_0.default,
                 fwhm=fwhm.default, bbox_factor=5.5, **kwargs):
        super().__init__(flux=flux, x_0=x_0, y_0=y_0, fwhm=fwhm, **kwargs)
        self.bbox_factor = bbox_factor

    @property
    def amplitude(self):
        """
        The peak amplitude of the Gaussian.
        """
        return _gaussian_amplitude(self.flux, self.sigma, self.sigma)

    @property
    def sigma(self):
        """
        Gaussian sigma (standard deviation).
        """
        return self.fwhm * GAUSSIAN_FWHM_TO_SIGMA

    def _calc_bounding_box(self, *, factor=5.5):
        """
        Calculate a bounding box defining the limits of the model.

        Parameters
        ----------
        factor : float, optional
            The multiple of the standard deviations (sigma) used to
            define the limits.

        Returns
        -------
        bbox : tuple
            A bounding box defining the ((y_min, y_max), (x_min, x_max))
            limits of the model.
        """
        delta = factor * self.sigma
        return ((self.y_0 - delta, self.y_0 + delta),
                (self.x_0 - delta, self.x_0 + delta))

    @property
    def bounding_box(self):
        """
        The bounding box of the model.

        Examples
        --------
        >>> from photutils.psf import CircularGaussianPRF
        >>> model = CircularGaussianPRF(x_0=0, y_0=0, fwhm=2)
        >>> model.bounding_box  # doctest: +FLOAT_CMP
        ModelBoundingBox(
            intervals={
                x: Interval(lower=-4.671269901584105, upper=4.671269901584105)
                y: Interval(lower=-4.671269901584105, upper=4.671269901584105)
            }
            model=CircularGaussianPRF(inputs=('x', 'y'))
            order='C'
        )
        >>> model.bbox_factor = 7
        >>> model.bounding_box  # doctest: +FLOAT_CMP
        ModelBoundingBox(
            intervals={
                x: Interval(lower=-5.945252602016134, upper=5.945252602016134)
                y: Interval(lower=-5.945252602016134, upper=5.945252602016134)
            }
            model=CircularGaussianPRF(inputs=('x', 'y'))
            order='C'
        )
        """
        return self._calc_bounding_box(factor=self.bbox_factor)

    def evaluate(self, x, y, flux, x_0, y_0, fwhm):
        """
        Calculate the value of the 2D Gaussian model at the input
        coordinates for the given model parameters.

        Parameters
        ----------
        x, y : float or array_like
            The x and y coordinates at which to evaluate the model.

        flux : float
            Total integrated flux over the entire PSF.

        x_0, y_0 : float
            Position of the peak along the x and y axes.

        fwhm : float
            FWHM of the Gaussian.

        Returns
        -------
        result : `~numpy.ndarray`
            The value of the model evaluated at the input coordinates.
        """
        x0 = x - x_0
        y0 = y - y_0
        sigma = fwhm * GAUSSIAN_FWHM_TO_SIGMA

        dpix = 0.5
        if isinstance(x0, u.Quantity):
            dpix <<= x0.unit

        return (flux / 4.0
                * ((erf((x0 + dpix) / (np.sqrt(2) * sigma))
                    - erf((x0 - dpix) / (np.sqrt(2) * sigma)))
                   * (erf((y0 + dpix) / (np.sqrt(2) * sigma))
                      - erf((y0 - dpix) / (np.sqrt(2) * sigma)))))

    @property
    def input_units(self):
        """
        The input units of the model.
        """
        x_unit = self.x_0.input_unit
        y_unit = self.y_0.input_unit
        if x_unit is None and y_unit is None:
            return None

        return {self.inputs[0]: x_unit, self.inputs[1]: y_unit}

    def _parameter_units_for_data_units(self, inputs_unit, outputs_unit):
        return {'x_0': inputs_unit[self.inputs[0]],
                'y_0': inputs_unit[self.inputs[0]],
                'fwhm': inputs_unit[self.inputs[0]],
                'flux': outputs_unit[self.outputs[0]]}


class CircularGaussianSigmaPRF(Fittable2DModel):
    r"""
    A circular 2D Gaussian PSF model integrated over pixels.

    This model is evaluated by integrating the 2D Gaussian over the
    input coordinate pixels, and is equivalent to assuming the PSF is
    2D Gaussian at a *sub-pixel* level. Because it is integrated over
    pixels, this model is considered a PRF instead of a PSF.

    The Gaussian is normalized such that the analytical integral over
    the entire 2D plane is equal to the total flux.

    This model is equivalent to `CircularGaussianPRF`, but it is
    parameterized in terms of the standard deviation (sigma) instead of
    the full width at half maximum (FWHM).

    Parameters
    ----------
    flux : float, optional
        Total integrated flux over the entire PSF.

    x_0 : float, optional
        Position of the peak in x direction.

    y_0 : float, optional
        Position of the peak in y direction.

    sigma : float, optional
        Width of the Gaussian PSF.

    bbox_factor : float, optional
        The multiple of the standard deviation (sigma) used to define
        the bounding box limits.

    **kwargs : dict, optional
        Additional optional keyword arguments to be passed to the
        `astropy.modeling.Model` parent class.

    See Also
    --------
    GaussianPSF, GaussianPRF, CircularGaussianPSF, CircularGaussianPRF

    Notes
    -----
    The circular Gaussian function is defined as:

    .. math::

        f(x, y) =
            \frac{F}{4}
            \left[
                {\rm erf} \left(\frac{x - x_0 + 0.5}
                                     {\sqrt{2} \sigma} \right)  -
                {\rm erf} \left(\frac{x - x_0 - 0.5}
                                     {\sqrt{2} \sigma} \right)
            \right]
            \left[
                {\rm erf} \left(\frac{y - y_0 + 0.5}
                                     {\sqrt{2} \sigma} \right) -
                {\rm erf} \left(\frac{y - y_0 - 0.5}
                                     {\sqrt{2} \sigma} \right)
            \right]

    where :math:`F` is the total integrated flux, :math:`(x_{0},
    y_{0})` is the position of the peak, :math:`\sigma` is the standard
    deviation of the Gaussian, and :math:`{\rm erf}` denotes the error
    function.

    The model is normalized such that:

    .. math::

        \int_{-\infty}^{\infty} \int_{-\infty}^{\infty} f(x, y) \,dx \,dy = F

    The ``sigma`` parameter is fixed by default. If you wish to fit this
    parameter, set the ``fixed`` attribute to `False`, e.g.,::

        >>> from photutils.psf import CircularGaussianSigmaPRF
        >>> model = CircularGaussianSigmaPRF()
        >>> model.sigma.fixed = False

    By default, the ``sigma`` parameter is bounded to be strictly
    positive.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Gaussian_function

    Examples
    --------
    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        import numpy as np
        from photutils.psf import CircularGaussianSigmaPRF

        model = CircularGaussianSigmaPRF(flux=71.4, x_0=24.3, y_0=25.2,
                                         sigma=5.1)
        yy, xx = np.mgrid[0:51, 0:51]
        data = model(xx, yy)
        fig, ax = plt.subplots()
        ax.imshow(data, origin='lower')
    """

    flux = Parameter(
        default=1, description='Total integrated flux over the entire PSF.')
    x_0 = Parameter(
        default=0, description='Position of the peak along the x axis')
    y_0 = Parameter(
        default=0, description='Position of the peak along the y axis')
    sigma = Parameter(
        default=1,
        bounds=(FLOAT_EPSILON, None),
        fixed=True,
        description='Sigma (standard deviation) of the Gaussian')

    def __init__(self, *, flux=flux.default, x_0=x_0.default, y_0=y_0.default,
                 sigma=sigma.default, bbox_factor=5.5, **kwargs):
        super().__init__(sigma=sigma, x_0=x_0, y_0=y_0, flux=flux, **kwargs)
        self.bbox_factor = bbox_factor

    @property
    def amplitude(self):
        """
        The peak amplitude of the Gaussian.
        """
        return _gaussian_amplitude(self.flux, self.sigma, self.sigma)

    @property
    def fwhm(self):
        """
        Gaussian FWHM.
        """
        return self.sigma / GAUSSIAN_FWHM_TO_SIGMA

    def _calc_bounding_box(self, *, factor=5.5):
        """
        Calculate a bounding box defining the limits of the model.

        Parameters
        ----------
        factor : float, optional
            The multiple of the standard deviations (sigma) used to
            define the limits.

        Returns
        -------
        bbox : tuple
            A bounding box defining the ((y_min, y_max), (x_min, x_max))
            limits of the model.
        """
        delta = factor * self.sigma
        return ((self.y_0 - delta, self.y_0 + delta),
                (self.x_0 - delta, self.x_0 + delta))

    @property
    def bounding_box(self):
        """
        The bounding box of the model.

        Examples
        --------
        >>> from photutils.psf import CircularGaussianPRF
        >>> model = CircularGaussianPRF(x_0=0, y_0=0, fwhm=2)
        >>> model.bounding_box  # doctest: +FLOAT_CMP
        ModelBoundingBox(
            intervals={
                x: Interval(lower=-4.671269901584105, upper=4.671269901584105)
                y: Interval(lower=-4.671269901584105, upper=4.671269901584105)
            }
            model=CircularGaussianPRF(inputs=('x', 'y'))
            order='C'
        )
        >>> model.bbox_factor = 7
        >>> model.bounding_box  # doctest: +FLOAT_CMP
        ModelBoundingBox(
            intervals={
                x: Interval(lower=-5.945252602016134, upper=5.945252602016134)
                y: Interval(lower=-5.945252602016134, upper=5.945252602016134)
            }
            model=CircularGaussianPRF(inputs=('x', 'y'))
            order='C'
        )
        """
        return self._calc_bounding_box(factor=self.bbox_factor)

    def evaluate(self, x, y, flux, x_0, y_0, sigma):
        """
        Calculate the value of the 2D Gaussian model at the input
        coordinates for the given model parameters.

        Parameters
        ----------
        x, y : float or array_like
            The coordinates at which to evaluate the model.

        flux : float
            The total flux of the star.

        x_0, y_0 : float
            The position of the star.

        sigma : float
            The width of the Gaussian PRF.

        Returns
        -------
        evaluated_model : `~numpy.ndarray`
            The evaluated model.
        """
        dpix = 0.5
        if isinstance(x_0, u.Quantity):
            dpix *= x_0.unit

        return (flux / 4
                * ((erf((x - x_0 + dpix) / (np.sqrt(2) * sigma))
                    - erf((x - x_0 - dpix) / (np.sqrt(2) * sigma)))
                   * (erf((y - y_0 + dpix) / (np.sqrt(2) * sigma))
                      - erf((y - y_0 - dpix) / (np.sqrt(2) * sigma)))))

    @property
    def input_units(self):
        """
        The input units of the model.
        """
        x_unit = self.x_0.input_unit
        y_unit = self.y_0.input_unit
        if x_unit is None and y_unit is None:
            return None

        return {self.inputs[0]: x_unit, self.inputs[1]: y_unit}

    def _parameter_units_for_data_units(self, inputs_unit, outputs_unit):
        # Note that here we need to make sure that x and y are in the same
        # units otherwise this can lead to issues since rotation is not well
        # defined.
        if inputs_unit[self.inputs[0]] != inputs_unit[self.inputs[1]]:
            msg = "Units of 'x' and 'y' inputs should match"
            raise UnitsError(msg)

        return {'x_0': inputs_unit[self.inputs[0]],
                'y_0': inputs_unit[self.inputs[0]],
                'sigma': inputs_unit[self.inputs[0]],
                'flux': outputs_unit[self.outputs[0]]}


class MoffatPSF(Fittable2DModel):
    r"""
    A 2D Moffat PSF model.

    This model is evaluated by sampling the 2D Moffat function at the
    input coordinates. The Moffat profile is normalized such that the
    analytical integral over the entire 2D plane is equal to the total
    flux.

    Parameters
    ----------
    flux : float, optional
        Total integrated flux over the entire PSF.

    x_0 : float, optional
        Position of the peak along the x-axis.

    y_0 : float, optional
        Position of the peak along the y-axis.

    alpha : float, optional
        The characteristic radius of the Moffat profile.

    beta : float, optional
        The asymptotic power-law slope of the Moffat profile wings at
        large radial distances. Larger values provide less flux in the
        profile wings. For large ``beta``, this profile approaches a
        Gaussian profile. ``beta`` must be greater than 1. If ``beta``
        is set to 1, then the Moffat profile is a Lorentz function,
        whose integral is infinite. For this normalized model, if
        ``beta`` is set to 1, then the profile will be zero everywhere.

    bbox_factor : float, optional
        The multiple of the FWHM used to define the bounding box limits.

    **kwargs : dict, optional
        Additional optional keyword arguments to be passed to the
        `astropy.modeling.Model` base class.

    See Also
    --------
    GaussianPSF, CircularGaussianPSF, GaussianPRF, CircularGaussianPRF

    Notes
    -----
    The Moffat profile is defined as:

    .. math::

       f(x, y) = F \frac{\beta - 1}{\pi \alpha^2}
           \left(1 + \frac{\left(x - x_{0}\right)^{2}
               + \left(y - y_{0}\right)^{2}}{\alpha^{2}}\right)^{-\beta}

    where :math:`F` is the total integrated flux and :math:`(x_{0},
    y_{0})` is the position of the peak. Note that :math:`\beta` must be
    greater than 1.

    The FWHM of the Moffat profile is given by:

    .. math::

        \rm{FWHM} = 2 \alpha \sqrt{2^{1 / \beta} - 1}

    The model is normalized such that, for :math:`\beta > 1`:

    .. math::

        \int_{-\infty}^{\infty} \int_{-\infty}^{\infty} f(x, y)
            \,dx \,dy = F

    The ``alpha`` and ``beta`` parameters are fixed by default. If
    you wish to fit these parameters, set the ``fixed`` attribute to
    `False`, e.g.,::

        >>> from photutils.psf import MoffatPSF
        >>> model = MoffatPSF()
        >>> model.alpha.fixed = False
        >>> model.beta.fixed = False

    By default, the ``alpha`` parameter is bounded to be strictly
    positive and the ``beta`` parameter is bounded to be greater than 1.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Moffat_distribution

    .. [2] https://ui.adsabs.harvard.edu/abs/1969A%26A.....3..455M/abstract

    .. [3] https://ned.ipac.caltech.edu/level5/Stetson/Stetson2_2_1.html

    Examples
    --------
    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        import numpy as np
        from photutils.psf import MoffatPSF

        model = MoffatPSF(flux=71.4, x_0=24.3, y_0=25.2, alpha=5.1, beta=3.2)
        yy, xx = np.mgrid[0:51, 0:51]
        data = model(xx, yy)
        fig, ax = plt.subplots()
        ax.imshow(data, origin='lower')
    """

    flux = Parameter(
        default=1, description='Total integrated flux over the entire PSF.')
    x_0 = Parameter(
        default=0, description='Position of the peak along the x axis')
    y_0 = Parameter(
        default=0, description='Position of the peak along the y axis')
    alpha = Parameter(
        default=1,
        bounds=(FLOAT_EPSILON, None),
        fixed=True,
        description='Characteristic radius of the Moffat profile')
    beta = Parameter(
        default=2,
        bounds=(1.0 + FLOAT_EPSILON, None),
        fixed=True,
        description='Power-law index of the Moffat profile')

    def __init__(self, *, flux=flux.default, x_0=x_0.default, y_0=y_0.default,
                 alpha=alpha.default, beta=beta.default, bbox_factor=10.0,
                 **kwargs):
        super().__init__(flux=flux, x_0=x_0, y_0=y_0, alpha=alpha, beta=beta,
                         **kwargs)
        self.bbox_factor = bbox_factor

    @property
    def fwhm(self):
        """
        The FWHM of the Moffat profile.
        """
        return 2.0 * self.alpha * np.sqrt(2 ** (1.0 / self.beta) - 1)

    def _calc_bounding_box(self, *, factor=10.0):
        """
        Calculate a bounding box defining the limits of the model.

        Parameters
        ----------
        factor : float, optional
            The multiple of the FWHM used to define the limits.

        Returns
        -------
        bbox : tuple
            A bounding box defining the ((y_min, y_max), (x_min, x_max))
            limits of the model.
        """
        delta = factor * self.fwhm
        return ((self.y_0 - delta, self.y_0 + delta),
                (self.x_0 - delta, self.x_0 + delta))

    @property
    def bounding_box(self):
        """
        The bounding box of the model.

        Examples
        --------
        >>> from photutils.psf import MoffatPSF
        >>> model = MoffatPSF(x_0=0, y_0=0, alpha=2, beta=3)
        >>> model.bounding_box  # doctest: +FLOAT_CMP
        ModelBoundingBox(
            intervals={
                x: Interval(lower=-20.39298114135835, upper=20.39298114135835)
                y: Interval(lower=-20.39298114135835, upper=20.39298114135835)
            }
            model=MoffatPSF(inputs=('x', 'y'))
            order='C'
        )
        >>> model.bbox_factor = 7
        >>> model.bounding_box  # doctest: +FLOAT_CMP
        ModelBoundingBox(
            intervals={
                x: Interval(lower=-14.27508679895084, upper=14.27508679895084)
                y: Interval(lower=-14.27508679895084, upper=14.27508679895084)
            }
            model=MoffatPSF(inputs=('x', 'y'))
            order='C'
        )
        """
        return self._calc_bounding_box(factor=self.bbox_factor)

    def evaluate(self, x, y, flux, x_0, y_0, alpha, beta):
        """
        Calculate the value of the 2D Moffat model at the input
        coordinates for the given model parameters.

        Parameters
        ----------
        x, y : float or array_like
            The x and y coordinates at which to evaluate the model.

        flux : float
            Total integrated flux over the entire PSF.

        x_0, y_0 : float
            Position of the peak along the x and y axes.

        alpha : float, optional
            The characteristic radius of the Moffat profile.

        beta : float, optional
            The asymptotic power-law slope of the Moffat profile wings
            at large radial distances. Larger values provide less flux
            in the profile wings. For large ``beta``, this profile
            approaches a Gaussian profile. ``beta`` must be greater
            than 1. If ``beta`` is set to 1, then the Moffat profile is
            a Lorentz function, whose integral is infinite. For this
            normalized model, if ``beta`` is set to 1, then the profile
            will be zero everywhere.

        Returns
        -------
        result : `~numpy.ndarray`
            The value of the model evaluated at the input coordinates.
        """
        # output units should match the input flux units
        alpha2 = alpha.copy()
        if isinstance(alpha, u.Quantity):
            alpha2 = alpha.value

        amp = flux * (beta - 1) / (np.pi * alpha2 ** 2)
        r2 = (x - x_0) ** 2 + (y - y_0) ** 2
        return amp * (1 + (r2 / alpha**2)) ** (-beta)

    @property
    def input_units(self):
        """
        The input units of the model.
        """
        x_unit = self.x_0.input_unit
        y_unit = self.y_0.input_unit
        if x_unit is None and y_unit is None:
            return None

        return {self.inputs[0]: x_unit, self.inputs[1]: y_unit}

    def _parameter_units_for_data_units(self, inputs_unit, outputs_unit):
        return {'x_0': inputs_unit[self.inputs[0]],
                'y_0': inputs_unit[self.inputs[0]],
                'alpha': inputs_unit[self.inputs[0]],
                'flux': outputs_unit[self.outputs[0]]}


class AiryDiskPSF(Fittable2DModel):
    r"""
    A 2D Airy disk PSF model.

    This model is evaluated by sampling the 2D Airy disk function at the
    input coordinates. The Airy disk profile is normalized such that the
    analytical integral over the entire 2D plane is equal to the total
    flux.

    Parameters
    ----------
    flux : float, optional
        Total integrated flux over the entire PSF.

    x_0 : float, optional
        Position of the peak along the x-axis.

    y_0 : float, optional
        Position of the peak along the y-axis.

    radius : float, optional
        The radius of the Airy disk at the first zero.

    bbox_factor : float, optional
        The multiple of the FWHM used to define the bounding box limits.

    **kwargs : dict, optional
        Additional optional keyword arguments to be passed to the
        `astropy.modeling.Model` base class.

    See Also
    --------
    GaussianPSF, CircularGaussianPSF, MoffatPSF

    Notes
    -----
    The Airy disk profile is defined as:

    .. math::

        f(r) = \frac{F}{4 \pi (R / R_z)^2}
               \left[ \frac{2 J_1\left(\frac{\pi r}{R / R_z}\right)}
                      {\frac{\pi r}{R / R_z}} \right]^2

    where :math:`r` is radial distance from the peak

    .. math::

        r = \sqrt{(x - x_0)^2 + (y - y_0)^2}

    :math:`F` is the total integrated flux,
    :math:`J_1` is the first order `Bessel function
    <https://en.wikipedia.org/wiki/Bessel_function>`_ of the first
    kind, :math:`R` is the input ``radius`` parameter, and :math:`R_z =
    1.2196698912665045` is the solution to the equation :math:`J_1(\pi
    R_z) = 0`.

    For an optical system, the radius of the first zero represents
    the limiting angular resolution. The limiting angular resolution
    is :math:`R_z \, \lambda / D \approx 1.22 \, \lambda / D`, where
    :math:`\lambda` is the wavelength of the light and :math:`D` is the
    diameter of the aperture.

    The full width at half maximum (FWHM) of the Airy disk profile is
    given by:

    .. math::

        \rm{FWHM} = 1.028993969962188 \, \frac{R}{R_z}
                  = 0.8436659602162364 \, R

    The model is normalized such that:

    .. math::

        \int_{0}^{2 \pi} \int_{0}^{\infty} f(r) \,r \,dr \,d\theta =
        \int_{-\infty}^{\infty} \int_{-\infty}^{\infty} f(x, y)
            \,dx \,dy = F

    The ``radius`` parameter is fixed by default. If you wish to fit
    this parameter, set the ``fixed`` attribute to `False`, e.g.,::

        >>> from photutils.psf import AiryDiskPSF
        >>> model = AiryDiskPSF()
        >>> model.radius.fixed = False

    By default, the ``radius`` parameter is bounded to be strictly
    positive.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Airy_disk

    Examples
    --------
    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        import numpy as np
        from astropy.visualization import simple_norm
        from photutils.psf import AiryDiskPSF

        model = AiryDiskPSF(flux=71.4, x_0=24.3, y_0=25.2, radius=5)
        yy, xx = np.mgrid[0:51, 0:51]
        data = model(xx, yy)
        norm = simple_norm(data, 'sqrt')
        fig, ax = plt.subplots()
        ax.imshow(data, norm=norm, origin='lower')
    """

    flux = Parameter(
        default=1, description='Total integrated flux over the entire PSF.')
    x_0 = Parameter(
        default=0, description='Position of the peak along the x axis')
    y_0 = Parameter(
        default=0, description='Position of the peak along the y axis')
    radius = Parameter(
        default=1,
        bounds=(FLOAT_EPSILON, None),
        fixed=True,
        description='Radius of the Airy disk at the first zero')

    _rz = jn_zeros(1, 1)[0] / np.pi

    def __init__(self, *, flux=flux.default, x_0=x_0.default, y_0=y_0.default,
                 radius=radius.default, bbox_factor=10.0, **kwargs):
        super().__init__(flux=flux, x_0=x_0, y_0=y_0, radius=radius, **kwargs)
        self.bbox_factor = bbox_factor

    @property
    def fwhm(self):
        """
        The FWHM of the Airy disk profile.
        """
        return 2.0 * 1.616339948310703 * self.radius / self._rz / np.pi

    def _calc_bounding_box(self, *, factor=10.0):
        """
        Calculate a bounding box defining the limits of the model.

        Parameters
        ----------
        factor : float, optional
            The multiple of the FWHM used to define the limits.

        Returns
        -------
        bbox : tuple
            A bounding box defining the ((y_min, y_max), (x_min, x_max))
            limits of the model.
        """
        delta = factor * self.fwhm
        return ((self.y_0 - delta, self.y_0 + delta),
                (self.x_0 - delta, self.x_0 + delta))

    @property
    def bounding_box(self):
        """
        The bounding box of the model.

        Examples
        --------
        >>> from photutils.psf import AiryDiskPSF
        >>> model = AiryDiskPSF(x_0=0, y_0=0, radius=3)
        >>> model.bounding_box  # doctest: +FLOAT_CMP
        ModelBoundingBox(
            intervals={
                x: Interval(lower=-25.30997880648709, upper=25.30997880648709)
                y: Interval(lower=-25.30997880648709, upper=25.30997880648709)
            }
            model=AiryDiskPSF(inputs=('x', 'y'))
            order='C'
        )
        >>> model.bbox_factor = 7
        >>> model.bounding_box  # doctest: +FLOAT_CMP
        ModelBoundingBox(
            intervals={
                x: Interval(lower=-17.71698516454096, upper=17.71698516454096)
                y: Interval(lower=-17.71698516454096, upper=17.71698516454096)
            }
            model=AiryDiskPSF(inputs=('x', 'y'))
            order='C'
        )
        """
        return self._calc_bounding_box(factor=self.bbox_factor)

    def evaluate(self, x, y, flux, x_0, y_0, radius):
        """
        Calculate the value of the 2D Airy disk model at the input
        coordinates for the given model parameters.

        Parameters
        ----------
        x, y : float or array_like
            The x and y coordinates at which to evaluate the model.

        flux : float
            Total integrated flux over the entire PSF.

        x_0, y_0 : float
            Position of the peak along the x and y axes.

        radius : float, optional
            The radius of the Airy disk at the first zero.

        Returns
        -------
        result : `~numpy.ndarray`
            The value of the model evaluated at the input coordinates.
        """
        r = np.sqrt((x - x_0) ** 2 + (y - y_0) ** 2) / (radius / self._rz)

        if isinstance(r, u.Quantity):
            # scipy function cannot handle Quantity, so turn into array
            r = r.to_value(u.dimensionless_unscaled)

        # Since r can be zero, we have to take care to treat that case
        # separately so as not to raise a numpy warning
        z = np.ones(r.shape)
        rt = np.pi * r[r > 0]
        z[r > 0] = (2.0 * j1(rt) / rt) ** 2

        if isinstance(flux, u.Quantity):
            # make z a quantity to allow in-place multiplication
            z <<= u.dimensionless_unscaled

        normalization = (4.0 / np.pi) * (radius / self._rz) ** 2
        if isinstance(normalization, u.Quantity):
            normalization = normalization.value

        z *= (flux / normalization)

        return z

    @property
    def input_units(self):
        """
        The input units of the model.
        """
        x_unit = self.x_0.input_unit
        y_unit = self.y_0.input_unit
        if x_unit is None and y_unit is None:
            return None

        return {self.inputs[0]: x_unit, self.inputs[1]: y_unit}

    def _parameter_units_for_data_units(self, inputs_unit, outputs_unit):
        return {'x_0': inputs_unit[self.inputs[0]],
                'y_0': inputs_unit[self.inputs[0]],
                'radius': inputs_unit[self.inputs[0]],
                'flux': outputs_unit[self.outputs[0]]}



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the _deprecation module.
"""

import warnings

import numpy as np
import pytest
from astropy.table import QTable, Table, TableMergeError, join, unique
from astropy.utils.exceptions import AstropyDeprecationWarning

from photutils.utils._deprecation import (DeprecatedColumnQTable,
                                          DeprecatedColumnTable,
                                          _future_column_names_var,
                                          create_deprecated_table_from_data,
                                          create_empty_deprecated_qtable,
                                          deprecated, deprecated_getattr,
                                          deprecated_positional_kwargs,
                                          deprecated_renamed_argument,
                                          use_future_column_names)

DEPRECATION_MAP = {'old': 'new', 'old_b': 'new_b'}


@pytest.fixture
def raw_data():
    """
    Provide a raw data dictionary for table creation.
    """
    return {'old': [3, 2, 1], 'old_b': [4, 5, 6], 'stable': [7, 8, 9]}


class TestDeprecatedColumn:
    """
    Tests for DeprecatedColumnTable and DeprecatedColumnQTable.
    """

    def test_creation_and_type(self, raw_data):
        """
        Test that the factory creates the correct object type.
        """
        table = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        assert isinstance(table, DeprecatedColumnTable)
        assert not isinstance(table, DeprecatedColumnQTable)
        assert set(table.colnames) == {'new', 'new_b', 'stable'}

        qtable = create_deprecated_table_from_data(
            raw_data, DEPRECATION_MAP, use_qtable=True)
        assert isinstance(qtable, DeprecatedColumnQTable)

    def test_masked_creation(self, raw_data):
        """
        Test that kwargs like "masked" are passed through correctly.
        """
        table = create_deprecated_table_from_data(
            raw_data, DEPRECATION_MAP, masked=True,
        )
        assert isinstance(table, DeprecatedColumnTable)
        assert table.masked is True
        table['new'].mask[0] = True
        assert np.all(table['new'].mask == [True, False, False])

    def test_getitem_access(self, raw_data):
        """
        Test deprecated access via __getitem__.
        """
        table = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)

        match = "'old' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            col = table['old']
        assert np.all(col == table['new'])

        match = "'old_b' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            sub_table = table[['stable', 'old_b']]
        assert sub_table.colnames == ['stable', 'new_b']

    def test_setitem_assignment(self, raw_data):
        """
        Test deprecated assignment via __setitem__.
        """
        table = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        match = "'old' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            table['old'] = [100, 200, 300]
        assert np.all(table['new'] == [100, 200, 300])

    def test_delitem_and_remove(self, raw_data):
        """
        Test deprecated deletion via __delitem__ and remove methods.
        """
        table1 = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        match = "'old' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            del table1['old']
        assert 'new' not in table1.colnames

        table2 = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        with pytest.warns(AstropyDeprecationWarning, match=match):
            table2.remove_column('old')
        assert 'new' not in table2.colnames

    def test_keep_columns(self, raw_data):
        """
        Test deprecated use in keep_columns.
        """
        table = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        match = "'old' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            table.keep_columns(['stable', 'old'])
        assert set(table.colnames) == {'stable', 'new'}

    def test_rename_methods(self, raw_data):
        """
        Test deprecated use in rename_column and rename_columns.
        """
        table1 = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        match = "'old' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            table1.rename_column('old', 'final_name_1')
        assert 'final_name_1' in table1.colnames
        assert 'new' not in table1.colnames

        table2 = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        with pytest.warns(AstropyDeprecationWarning):
            table2.rename_columns(['old', 'old_b'], ['final1', 'final2'])
        assert set(table2.colnames) == {'final1', 'final2', 'stable'}

    def test_data_operations(self, raw_data):
        """
        Test deprecated use in sort, group_by, and unique.
        """
        table_sort = create_deprecated_table_from_data(raw_data,
                                                       DEPRECATION_MAP)
        match = "'old' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            table_sort.sort('old')
        assert table_sort['new'][0] == 1

        table_group = create_deprecated_table_from_data(raw_data,
                                                        DEPRECATION_MAP)
        match = "'old_b' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            groups = table_group.group_by('old_b')
        assert len(groups.groups) == 3

        table_unique = create_deprecated_table_from_data(raw_data,
                                                         DEPRECATION_MAP)
        match = "'old' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            unique_table = unique(table_unique, keys='old')
        assert len(unique_table) == 3

    def test_join(self, raw_data, recwarn):
        """
        Test deprecated use in the standalone join function.
        """
        table1 = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        table2 = Table({'new': [1, 3], 'extra': [1.1, 3.3]})

        match = "Left table does not have key column 'old'"
        with pytest.raises(TableMergeError, match=match):
            join(table1, table2, keys='old')

        # Test that it works correctly with the new name and issues no
        # warnings
        joined = join(table1, table2, keys='new')
        assert 'extra' in joined.colnames
        assert len(joined) == 2
        assert len(recwarn) == 0

    def test_indexing(self, raw_data, recwarn):
        """
        Test deprecated use in add_index and remove_indices.
        """
        table = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        match = "'old' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            table.add_index('old')

        assert len(table.indices) == 1
        assert table.indices[0].columns[0].name == 'new'

        # The `pytest.warns` context manager consumes the warning. Now we can
        # test that the next operation issues no new warnings.
        table.remove_indices('new')
        assert not table.indices
        assert len(recwarn) == 0

    def test_non_string_access_no_warning(self, raw_data, recwarn):
        """
        Test that non-string access does not trigger warnings.

        This ensures that row access via integers or slices does not
        incorrectly engage the name translation logic.
        """
        table = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)

        # Access a row by integer
        row = table[0]
        assert row['new'] == 3

        # Slice rows
        sliced = table[0:2]
        assert len(sliced) == 2

        # Assert that no warnings were recorded during these operations
        assert len(recwarn) == 0

    def test_contains(self, raw_data):
        """
        Test deprecated use in ``in`` checks.
        """
        table = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        match = "'old' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            assert 'old' in table

        # Non-deprecated column: no warning
        assert 'stable' in table

        # Missing column: no warning
        assert 'missing' not in table

    def test_copy_preserves_deprecation(self, raw_data):
        """
        Test that copy() preserves deprecation behavior.
        """
        table = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        copied = table.copy()

        assert isinstance(copied, DeprecatedColumnTable)
        match = "'old' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            col = copied['old']
        assert np.all(col == [3, 2, 1])

        # Ensure it's a true copy; modifying original doesn't affect copy
        table['new'][0] = 999
        assert copied['new'][0] == 3

    def test_remove_columns(self, raw_data):
        """
        Test deprecated use in remove_columns (plural).
        """
        table = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        with pytest.warns(AstropyDeprecationWarning):
            table.remove_columns(['old', 'old_b'])
        assert table.colnames == ['stable']

    def test_replace_column(self, raw_data):
        """
        Test deprecated use in replace_column.
        """
        table = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        match = "'old' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            table.replace_column('old', [100, 200, 300])
        assert np.all(table['new'] == [100, 200, 300])

    def test_empty_deprecation_map(self, raw_data, recwarn):
        """
        Test a table with an empty deprecation map (no deprecated names).
        """
        table = create_deprecated_table_from_data(raw_data, {})  # empty map
        # All operations should work without any warnings
        assert set(table.colnames) == {'old', 'old_b', 'stable'}
        col = table['old']
        assert np.all(col == [3, 2, 1])
        table.sort('old')
        assert 'old' in table
        assert len(recwarn) == 0

    def test_add_index_list(self, raw_data):
        """
        Test deprecated use in add_index with a list of column names.
        """
        table = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        with pytest.warns(AstropyDeprecationWarning):
            table.add_index(['old', 'old_b'])
        assert len(table.indices) == 1
        index_col_names = [c.name for c in table.indices[0].columns]
        assert index_col_names == ['new', 'new_b']

    def test_remove_indices_deprecated(self, raw_data):
        """
        Test deprecated use in remove_indices.
        """
        table = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        table.add_index('new')
        assert len(table.indices) == 1
        with pytest.warns(AstropyDeprecationWarning):
            table.remove_indices('old')
        assert len(table.indices) == 0

    def test_copy_with_none_deprecation_map(self):
        """
        Test that copy() works when deprecation_map is None.
        """
        table = DeprecatedColumnTable({'a': [1, 2, 3]})
        table.deprecation_map = None
        copied = table.copy()
        assert copied.deprecation_map is None
        assert np.all(copied['a'] == [1, 2, 3])

    def test_create_empty_deprecated_qtable(self):
        """
        Test creating an empty QTable and adding columns incrementally.
        """
        table = create_empty_deprecated_qtable(DEPRECATION_MAP)
        assert isinstance(table, DeprecatedColumnQTable)
        assert len(table) == 0

        # Add columns using new names
        table['new'] = [1, 2, 3]
        table['stable'] = [4, 5, 6]

        # Access via deprecated name
        match = "'old' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            col = table['old']
        assert np.all(col == [1, 2, 3])

    def test_slice_preserves_deprecation(self, raw_data):
        """
        Test that slicing preserves deprecation behavior.
        """
        table = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)
        sliced = table[0:2]

        assert isinstance(sliced, DeprecatedColumnTable)
        match = "'old' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            col = sliced['old']
        assert np.all(col == [3, 2])

    def test_translate_names_non_string(self, raw_data, recwarn):
        """
        Test that _translate_names passes through non-string/non-sequence
        values unchanged (e.g., a Table used as group_by keys).
        """
        table = create_deprecated_table_from_data(raw_data, DEPRECATION_MAP)

        # group_by accepts a Table as keys; _translate_names should pass
        # it through without modification or warnings
        key_table = Table({'new': [3, 2, 1]})
        groups = table.group_by(key_table)
        assert len(groups.groups) == 3
        assert len(recwarn) == 0


class TestFutureColumnNames:
    """
    Tests for the ``photutils.future_column_names`` opt-in flag.
    """

    def setup_method(self):
        import photutils
        self._original = photutils.future_column_names
        photutils.future_column_names = True

    def teardown_method(self):
        import photutils
        photutils.future_column_names = self._original

    def test_from_data_returns_plain_table(self, raw_data):
        """
        Test that create_deprecated_table_from_data returns a plain
        Table when the flag is set.
        """
        table = create_deprecated_table_from_data(
            raw_data, DEPRECATION_MAP)
        assert type(table) is Table
        assert set(table.colnames) == {'new', 'new_b', 'stable'}

    def test_from_data_returns_plain_qtable(self, raw_data):
        """
        Test that create_deprecated_table_from_data returns a plain
        QTable when the flag is set and use_qtable=True.
        """
        table = create_deprecated_table_from_data(
            raw_data, DEPRECATION_MAP, use_qtable=True)
        assert type(table) is QTable

    def test_from_data_no_warnings(self, raw_data, recwarn):
        """
        Test that accessing columns on a plain table created with the
        flag does not issue deprecation warnings.
        """
        table = create_deprecated_table_from_data(
            raw_data, DEPRECATION_MAP)
        _ = table['new']
        assert len(recwarn) == 0

    def test_from_data_old_name_raises(self, raw_data):
        """
        Test that accessing a deprecated name on a plain table created
        with the flag raises KeyError.
        """
        table = create_deprecated_table_from_data(
            raw_data, DEPRECATION_MAP)
        with pytest.raises(KeyError):
            table['old']

    def test_empty_returns_plain_qtable(self):
        """
        Test that create_empty_deprecated_qtable returns a plain QTable
        when the flag is set.
        """
        table = create_empty_deprecated_qtable(DEPRECATION_MAP)
        assert type(table) is QTable
        assert len(table) == 0


class TestUseFutureColumnNames:
    """
    Tests for the ``use_future_column_names`` context manager.
    """

    def test_context_manager_returns_plain_table(self, raw_data):
        """
        Test that the context manager makes
        create_deprecated_table_from_data return a plain Table.
        """
        with use_future_column_names():
            table = create_deprecated_table_from_data(
                raw_data, DEPRECATION_MAP)
        assert type(table) is Table
        assert set(table.colnames) == {'new', 'new_b', 'stable'}

    def test_context_manager_returns_plain_qtable(self, raw_data):
        """
        Test that the context manager makes
        create_deprecated_table_from_data return a plain QTable.
        """
        with use_future_column_names():
            table = create_deprecated_table_from_data(
                raw_data, DEPRECATION_MAP, use_qtable=True)
        assert type(table) is QTable

    def test_context_manager_empty_qtable(self):
        """
        Test that the context manager makes
        create_empty_deprecated_qtable return a plain QTable.
        """
        with use_future_column_names():
            table = create_empty_deprecated_qtable(DEPRECATION_MAP)
        assert type(table) is QTable
        assert len(table) == 0

    def test_global_unchanged_after_context(self):
        """
        Test that the global flag is unchanged after the context
        manager exits.
        """
        import photutils

        original = photutils.future_column_names
        with use_future_column_names():
            pass
        assert photutils.future_column_names == original

    def test_outside_context_uses_global(self, raw_data):
        """
        Test that outside the context manager, the global flag is
        respected and deprecated table behavior is used.
        """
        import photutils

        assert not photutils.future_column_names
        with use_future_column_names():
            pass
        # Outside the context, should use the deprecated table
        table = create_deprecated_table_from_data(
            raw_data, DEPRECATION_MAP)
        assert type(table) is DeprecatedColumnTable

    def test_nested_context_managers(self, raw_data):
        """
        Test that nested context managers work correctly with different
        values.
        """
        with use_future_column_names(enabled=True):
            table1 = create_deprecated_table_from_data(
                raw_data, DEPRECATION_MAP)
            assert type(table1) is Table

            with use_future_column_names(enabled=False):
                table2 = create_deprecated_table_from_data(
                    raw_data, DEPRECATION_MAP)
                assert type(table2) is DeprecatedColumnTable

            # Back to the outer context
            table3 = create_deprecated_table_from_data(
                raw_data, DEPRECATION_MAP)
            assert type(table3) is Table

    def test_context_manager_disabled(self, raw_data):
        """
        Test that use_future_column_names(enabled=False) forces the
        deprecated table behavior even if the global flag is True.
        """
        import photutils

        original = photutils.future_column_names
        try:
            photutils.future_column_names = True
            with use_future_column_names(enabled=False):
                table = create_deprecated_table_from_data(
                    raw_data, DEPRECATION_MAP)
                assert type(table) is DeprecatedColumnTable
        finally:
            photutils.future_column_names = original

    def test_restores_on_exception(self):
        """
        Test that the context manager restores state even if an
        exception occurs.
        """
        sentinel_before = _future_column_names_var.get()
        msg = 'test error'
        with use_future_column_names(), pytest.raises(ValueError, match=msg):
            raise ValueError(msg)
        assert _future_column_names_var.get() == sentinel_before


@deprecated_positional_kwargs('1.0', until='2.0')
def _example_func(a, b=10, c=20):
    """
    Example function for testing deprecated_positional_kwargs.
    """
    return a + b + c


class TestDeprecatedPositionalKwargs:
    """
    Tests for the deprecated_positional_kwargs decorator.
    """

    def test_no_warning_at_limit(self):
        with warnings.catch_warnings():
            warnings.simplefilter('error')
            result = _example_func(1)
        assert result == 31

    def test_no_warning_keyword_only(self):
        with warnings.catch_warnings():
            warnings.simplefilter('error')
            result = _example_func(1, b=5, c=3)
        assert result == 9

    def test_warns_when_exceeded(self):
        match = "'_example_func'"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            result = _example_func(1, 2)
        assert result == 23

    def test_warning_message_versions(self):
        with pytest.warns(AstropyDeprecationWarning) as record:
            _example_func(1, 2, 3)
        msg = str(record[0].message)
        assert '1.0' in msg
        assert '2.0' in msg

    def test_warning_names_single(self):
        with pytest.warns(AstropyDeprecationWarning) as record:
            _example_func(1, 2)
        msg = str(record[0].message)
        assert "Passing 'b' positionally" in msg
        assert "'c'" not in msg
        assert 'Pass it as a keyword argument' in msg
        assert 'b=...' in msg

    def test_warning_names_two(self):
        with pytest.warns(AstropyDeprecationWarning) as record:
            _example_func(1, 2, 3)
        msg = str(record[0].message)
        assert "'b' and 'c'" in msg
        assert 'Pass them as keyword arguments' in msg
        assert 'b=..., c=...' in msg

    def test_warning_names_three(self):
        @deprecated_positional_kwargs('1.0')
        def _func(a, b=1, c=2, d=3):
            return a + b + c + d

        with pytest.warns(AstropyDeprecationWarning) as record:
            _func(1, 2, 3, 4)
        msg = str(record[0].message)
        assert "'b', 'c', and 'd'" in msg
        assert 'Pass them as keyword arguments' in msg
        assert 'b=..., c=..., d=...' in msg

    def test_return_value_preserved(self):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            assert _example_func(5, 3, 2) == 10
        assert _example_func(5) == 35

    def test_preserves_metadata(self):
        assert _example_func.__name__ == '_example_func'
        assert 'Example function' in _example_func.__doc__

    def test_zero_positional(self):
        @deprecated_positional_kwargs('1.5', until='2.5')
        def _no_pos(x=0):
            return x

        with warnings.catch_warnings():
            warnings.simplefilter('error')
            result = _no_pos(x=42)
        assert result == 42

        with pytest.warns(AstropyDeprecationWarning) as record:
            result = _no_pos(42)
        assert result == 42
        msg = str(record[0].message)
        assert "'x'" in msg
        assert 'Pass it as a keyword argument' in msg
        assert 'x=...' in msg

    def test_no_until(self):
        @deprecated_positional_kwargs('3.0')
        def _func(a, b=10):
            return a + b

        with pytest.warns(AstropyDeprecationWarning) as record:
            result = _func(1, 2)
        assert result == 3
        msg = str(record[0].message)
        assert '3.0' in msg
        assert 'a future version' in msg
        assert "'b'" in msg
        assert 'b=...' in msg

    def test_until_keyword_only(self):
        match = 'takes 1 positional argument'
        with pytest.raises(TypeError, match=match):
            # until passed positionally
            deprecated_positional_kwargs('1.0', '2.0')

    def test_since_until_int(self):
        @deprecated_positional_kwargs(3, until=4)
        def _func(a, b=10):
            return a + b

        with pytest.warns(AstropyDeprecationWarning) as record:
            result = _func(1, 2)
        assert result == 3
        msg = str(record[0].message)
        assert '3' in msg
        assert '4' in msg

    def test_multiple_required_args(self):
        @deprecated_positional_kwargs('1.0')
        def _func(a, b, c=10):
            return a + b + c

        # Two required positional args should not warn
        with warnings.catch_warnings():
            warnings.simplefilter('error')
            result = _func(1, 2)
        assert result == 13

        # Third (optional) arg passed positionally should warn
        with pytest.warns(AstropyDeprecationWarning) as record:
            result = _func(1, 2, 3)
        assert result == 6
        msg = str(record[0].message)
        assert "'c'" in msg
        assert "Passing 'c'" in msg
        assert "'a'" not in msg
        assert "'b'" not in msg

    def test_positional_only_params(self):
        @deprecated_positional_kwargs('1.0')
        def _func(a, /, b=10):
            return a + b

        # Positional-only arg should not warn
        with warnings.catch_warnings():
            warnings.simplefilter('error')
            result = _func(1)
        assert result == 11

        # Optional arg passed positionally should warn
        with pytest.warns(AstropyDeprecationWarning) as record:
            result = _func(1, 2)
        assert result == 3
        msg = str(record[0].message)
        assert "'b'" in msg


@deprecated_renamed_argument('b', 'new', '1.0', until='2.0')
def _example_func2(a, new, c=20):
    """
    Example function for testing deprecated_renamed_argument.
    """
    return a + new + c


@deprecated_renamed_argument('b', 'new', '1.0', until=None)
def _example_func3(a, new, c=20):
    """
    Example function for testing deprecated_renamed_argument
    with no "until" version specified.
    """
    return a + new + c


def test_deprecated_renamed_argument():
    # Test that using the new name works without warnings
    result = _example_func2(1, new=5, c=3)
    assert result == 9

    # Test that using the old name issues a warning and still works
    with pytest.warns(AstropyDeprecationWarning) as record:
        result = _example_func2(1, b=5, c=3)
    assert result == 9
    msg = str(record[0].message)
    assert "'b' was deprecated" in msg
    assert "'new'" in msg
    assert 'version 2.0' in msg

    # Test that if until=None, the warning is issued but no end version
    # is mentioned
    with pytest.warns(AstropyDeprecationWarning) as record:
        result = _example_func3(1, b=5, c=3)
    assert result == 9
    msg = str(record[0].message)
    assert 'deprecated' in msg.lower()
    assert 'future version' in msg


def test_deprecated_renamed_argument_always_warns():
    # Test that the warning is issued on every call, not just the first
    # time from a given call site.
    with pytest.warns(AstropyDeprecationWarning):
        _example_func2(1, b=5)
    with pytest.warns(AstropyDeprecationWarning):
        _example_func2(1, b=5)


class TestColumnDeprecationUntil:
    """
    Tests for the ``until`` parameter in column deprecation.
    """

    def test_until_in_warning_message(self):
        """
        Test that the removal version appears in the warning message.
        """
        table = create_deprecated_table_from_data(
            {'old': [1]}, DEPRECATION_MAP, until='5.0')
        with pytest.warns(AstropyDeprecationWarning) as record:
            _ = table['old']
        msg = str(record[0].message)
        assert 'version 5.0' in msg

    def test_until_none(self):
        """
        Test that without ``until``, the message says "a future
        version".
        """
        table = create_deprecated_table_from_data(
            {'old': [1]}, DEPRECATION_MAP)
        with pytest.warns(AstropyDeprecationWarning) as record:
            _ = table['old']
        msg = str(record[0].message)
        assert 'a future version' in msg

    def test_future_column_names(self):
        """
        Test that the warning mentions ``future_column_names``.
        """
        table = create_deprecated_table_from_data(
            {'old': [1]}, DEPRECATION_MAP)
        with pytest.warns(AstropyDeprecationWarning) as record:
            _ = table['old']
        msg = str(record[0].message)
        assert 'future_column_names' in msg

    def test_until_preserved_on_copy(self):
        """
        Test that copy() preserves the ``until`` value.
        """
        table = create_deprecated_table_from_data(
            {'old': [1]}, DEPRECATION_MAP, until='5.0')
        copied = table.copy()
        with pytest.warns(AstropyDeprecationWarning) as record:
            _ = copied['old']
        msg = str(record[0].message)
        assert 'version 5.0' in msg

    def test_until_preserved_on_slice(self):
        """
        Test that slicing preserves the ``until`` value.
        """
        table = create_deprecated_table_from_data(
            {'old': [1, 2]}, DEPRECATION_MAP, until='5.0')
        sliced = table[0:1]
        with pytest.warns(AstropyDeprecationWarning) as record:
            _ = sliced['old']
        msg = str(record[0].message)
        assert 'version 5.0' in msg

    def test_empty_qtable_until(self):
        """
        Test that create_empty_deprecated_qtable passes ``until``.
        """
        table = create_empty_deprecated_qtable(
            DEPRECATION_MAP, until='6.0')
        table['new'] = [1, 2]
        with pytest.warns(AstropyDeprecationWarning) as record:
            _ = table['old']
        msg = str(record[0].message)
        assert 'version 6.0' in msg

    def test_since_in_warning_message(self):
        """
        Test that the deprecation version appears in the warning
        message.
        """
        table = create_deprecated_table_from_data(
            {'old': [1]}, DEPRECATION_MAP, since='3.0')
        with pytest.warns(AstropyDeprecationWarning) as record:
            _ = table['old']
        msg = str(record[0].message)
        assert 'in version 3.0' in msg

    def test_since_none(self):
        """
        Test that without ``since``, the message does not mention a
        deprecation version.
        """
        table = create_deprecated_table_from_data(
            {'old': [1]}, DEPRECATION_MAP)
        with pytest.warns(AstropyDeprecationWarning) as record:
            _ = table['old']
        msg = str(record[0].message)
        assert 'was deprecated.' in msg
        assert 'was deprecated in version' not in msg

    def test_since_and_until(self):
        """
        Test that both ``since`` and ``until`` appear in the warning.
        """
        table = create_deprecated_table_from_data(
            {'old': [1]}, DEPRECATION_MAP, since='3.0', until='4.0')
        with pytest.warns(AstropyDeprecationWarning) as record:
            _ = table['old']
        msg = str(record[0].message)
        assert 'in version 3.0' in msg
        assert 'version 4.0' in msg

    def test_since_preserved_on_copy(self):
        """
        Test that copy() preserves the ``since`` value.
        """
        table = create_deprecated_table_from_data(
            {'old': [1]}, DEPRECATION_MAP, since='3.0', until='4.0')
        copied = table.copy()
        with pytest.warns(AstropyDeprecationWarning) as record:
            _ = copied['old']
        msg = str(record[0].message)
        assert 'in version 3.0' in msg
        assert 'version 4.0' in msg

    def test_since_preserved_on_slice(self):
        """
        Test that slicing preserves the ``since`` value.
        """
        table = create_deprecated_table_from_data(
            {'old': [1, 2]}, DEPRECATION_MAP, since='3.0', until='4.0')
        sliced = table[0:1]
        with pytest.warns(AstropyDeprecationWarning) as record:
            _ = sliced['old']
        msg = str(record[0].message)
        assert 'in version 3.0' in msg

    def test_empty_qtable_since(self):
        """
        Test that create_empty_deprecated_qtable passes ``since``.
        """
        table = create_empty_deprecated_qtable(
            DEPRECATION_MAP, since='3.0', until='4.0')
        table['new'] = [1, 2]
        with pytest.warns(AstropyDeprecationWarning) as record:
            _ = table['old']
        msg = str(record[0].message)
        assert 'in version 3.0' in msg
        assert 'version 4.0' in msg


class _ExampleObj:
    """
    A helper class for testing ``deprecated_getattr``.
    """

    def __init__(self):
        self.new_attr = 42
        self._deprecated_attrs = {'old_attr': 'new_attr'}

    def __getattr__(self, name):
        return deprecated_getattr(self, name, self._deprecated_attrs)


class _ExampleObjSinceUntil:
    """
    A helper class for testing ``deprecated_getattr`` with since/until.
    """

    def __init__(self):
        self.new_attr = 42
        self._deprecated_attrs = {'old_attr': 'new_attr'}

    def __getattr__(self, name):
        return deprecated_getattr(self, name, self._deprecated_attrs,
                                  since='3.0', until='4.0')


class TestDeprecatedGetattr:
    """
    Tests for the ``deprecated_getattr`` helper function.
    """

    def test_deprecated_access_warns(self):
        """
        Test that accessing a deprecated attribute issues a warning.
        """
        obj = _ExampleObj()
        match = "'old_attr'.*deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            val = obj.old_attr
        assert val == 42

    def test_new_name_no_warning(self):
        """
        Test that the new attribute does not trigger a warning.
        """
        obj = _ExampleObj()
        assert obj.new_attr == 42

    def test_unknown_attr_raises(self):
        """
        Test that an unknown attribute raises AttributeError.
        """
        obj = _ExampleObj()
        match = 'no attribute'
        with pytest.raises(AttributeError, match=match):
            _ = obj.nonexistent

    def test_message_no_since_no_until(self):
        """
        Test the default message (no "since", no "until").
        """
        obj = _ExampleObj()
        with pytest.warns(AstropyDeprecationWarning) as record:
            _ = obj.old_attr
        msg = str(record[0].message)
        assert "'old_attr'" in msg
        assert "'new_attr'" in msg
        assert 'a future version' in msg
        assert 'in version' not in msg

    def test_message_with_since_and_until(self):
        """
        Test the message includes "since" and "until" versions.
        """
        obj = _ExampleObjSinceUntil()
        with pytest.warns(AstropyDeprecationWarning) as record:
            _ = obj.old_attr
        msg = str(record[0].message)
        assert 'in version 3.0' in msg
        assert 'version 4.0' in msg

    def test_message_with_since_only(self):
        """
        Test the message when only "since" is provided.
        """
        obj = _ExampleObj()
        dep_map = {'x': 'y'}
        obj.y = 99
        with pytest.warns(AstropyDeprecationWarning) as record:
            val = deprecated_getattr(obj, 'x', dep_map, since='2.0')
        assert val == 99
        msg = str(record[0].message)
        assert 'in version 2.0' in msg
        assert 'a future version' in msg

    def test_message_with_until_only(self):
        """
        Test the message when only "until" is provided.
        """
        obj = _ExampleObj()
        dep_map = {'x': 'y'}
        obj.y = 99
        with pytest.warns(AstropyDeprecationWarning) as record:
            val = deprecated_getattr(obj, 'x', dep_map, until='5.0')
        assert val == 99
        msg = str(record[0].message)
        assert 'version 5.0' in msg
        # since was not given, so "deprecated in version" should not appear
        assert 'deprecated in version' not in msg


@deprecated('1.0', until='2.0')
def _example_func4(a, b=10, c=20):
    """
    Example function for testing deprecated_positional_kwargs.
    """
    return a + b + c


@deprecated('1.0')
def _example_func5(a, b=10, c=20):
    """
    Example function for testing deprecated_positional_kwargs.
    """
    return a + b + c


def test_deprecated():
    """
    Test the basic functionality of the @deprecated decorator.
    """
    with pytest.warns(AstropyDeprecationWarning) as record:
        result = _example_func4(1, 2, 3)
    assert result == 6
    msg = str(record[0].message)
    assert 'version 1.0' in msg
    assert 'version 2.0' in msg

    with pytest.warns(AstropyDeprecationWarning) as record:
        result = _example_func5(1, 2, 3)
    assert result == 6
    msg = str(record[0].message)
    assert 'version 1.0' in msg
    assert 'a future version' in msg
